package jTrolog.engine;

import jTrolog.lib.BasicLibrary;
import jTrolog.errors.PrologException;
import jTrolog.parser.Parser;
import jTrolog.terms.*;
import jTrolog.terms.Number;
import jTrolog.terms.Double;

import java.util.*;

/**
 * Keeps track of what variables are bound to what Terms and at which ChoicePoint this connection was made.
 *
 * now, the BindingsTable never receives anything that is not Wrapped. All Structs sent in are wrapped as a mfck.
 * the unifyBranch is now possible to run without generating new objects when the getArg is called.
 *
 * all the methods that add vars and get vars from the bidnings table is now given both Var and context (and link term and its context)
 * a problem is that the stackMap needs to be subsumed into the bindings table since it needs to retract the links done by a current owner.
 * I don't think I can remove the current owner as a separate map since it provides a rollback function. What I can do is create WrapVars for it only and then
 * move these into and out of the bindingsMap
 *
 * @author janerist@stud.ntnu.no
 * @author ivar.orstavik@hist.no
 */
public class BindingsTable {

    private int currentOwner = 0;

    private LinkTable links = new LinkTable();

    int uniqueExecutionCtxID = -1;
    public int getUniqueExecutionCtxID() {
        return uniqueExecutionCtxID--;
    }

    static BindingsTable onlyForIteration = new BindingsTable();

    //todo 1. perhaps it would be smart to keep track of the left and right terms/ctxs that the unification process works on.
    //todo    if this is done, then whenever a unification fails, one could display exactly what two terms did not match.
    //todo    but this is perhaps better done in the Engine, since one here might like to view the entire terms..
    public boolean unify(Term t0, Term t1) {
            int oneCtx = t0 instanceof Wrapper ? ((Wrapper) t0).getContext() : 0;
            int twoCtx = t1 instanceof Wrapper ? ((Wrapper) t1).getContext() : 0;
            return unifyBranch(unWrap(t0), unWrap(t1), oneCtx, twoCtx);
    }

    /**
     * @param one
     * @param two
     * @param oneCtx
     * @param twoCtx
     * @return true if the two terms are unified, false otherwise
     */
    boolean unifyBranch(Term one, Term two, int oneCtx, int twoCtx) {
        if (one instanceof Var) {
            //if the two varaiables are equal, then return true
            if (oneCtx == twoCtx && one.equals(two))
                return true;
            //else, if the left variable is not bound, bind it
            Term link = links.getTerm((Var)one, oneCtx);
            if (link == null)
                return setLink((Var) one, oneCtx, two, twoCtx);
            //else, Var one is bound. try to unify two with whatever one is bound to
            int linkCtx = links.getCtx((Var)one, oneCtx);
            return unifyBranch(link, two, linkCtx, twoCtx);
        }
        if (two instanceof Var){
            Term link = links.getTerm((Var)two, twoCtx);
            if(link == null)
                return setLink((Var) two, twoCtx, one, oneCtx);           //var two is unbound, bind it against one

            int linkCtx = links.getCtx((Var)two, twoCtx);
            return unifyBranch(one, link, oneCtx, linkCtx);    //Var two is bound, unify against two's link
        }
        //from now on, neither branch is a Var
        if (one instanceof StructAtom && two instanceof StructAtom)
            return one.equals(two);
        if (one instanceof Number && two instanceof jTrolog.terms.Number)
            return BasicLibrary.number_equality_2((Number)one, (Number)two);

        //just one empty
        if(one == Struct.emptyList ^ two == Struct.emptyList)
            return false;
        //two empties
        if(one == Struct.emptyList /*&& two == StructList.empty*/)
            return true;

        if (one instanceof Struct && two instanceof Struct) {
            Struct a = (Struct) one;
            Struct b = (Struct) two;
            if (a.getPredicateIndicator() != b.getPredicateIndicator())
                return false;
            if (a.getPredicateIndicator() == Struct.listSignature)
                return unifyLists(a, b, oneCtx, twoCtx);
            for (int c = 0; c < a.getArity(); c++)
                if (!unifyBranch(a.getArg(c), b.getArg(c), oneCtx, twoCtx))
                    return false;
            return true;
        }
        return false;
    }

    private boolean unifyLists(Struct left, Struct right, int leftCtx, int rightCtx){
        Term leftTail;
        Term rightTail;
        do {
            if (!unifyBranch(left.getArg(0), right.getArg(0), leftCtx, rightCtx))
                return false;
            leftTail = left.getArg(1);
            rightTail = right.getArg(1);
            if (!(leftTail instanceof Struct) || !(rightTail instanceof Struct))
                break;
            left = (Struct) leftTail;
            right = (Struct) rightTail;
        } while (left.getPredicateIndicator() == right.getPredicateIndicator() &&
                left.getPredicateIndicator() == Struct.listSignature);
        return unifyBranch(leftTail, rightTail, leftCtx, rightCtx);
    }

    public int listSize(Struct list) {
        if (list == Struct.emptyList)
            return 0;
        return Struct.size(((StructList) list));
    }

    public Term resolveAndCloneStructs(Term t, int wrapperCtx){
        varPosNumberCounter = 1;
        return wrapWithID(resolveAndCloneStructsImpl(t), wrapperCtx);
    }

    private int varPosNumberCounter;
    private Term resolveAndCloneStructsImpl(Term t) {
        if (t instanceof Number || t instanceof StructAtom || t == Struct.emptyList)
            return t;
        if (t instanceof Var){
            if (!(t instanceof WrapVar))
                return t;
            Term lastLink = resolve(t);
            if (t == lastLink){
                int ctx = ((WrapVar) t).getContext();
                if (ctx == 0)
                    return ((WrapVar) t).getBasis();
                return new Var(t.toString() + "_" + ctx, varPosNumberCounter++);
            }
            return resolveAndCloneStructsImpl(lastLink);
        }
        if (t instanceof WrapStructList){
            LinkedList clonedElements = new LinkedList();
            for (Iterator it = structListIterator((Struct) t, false); it.hasNext();)
                clonedElements.add(resolveAndCloneStructsImpl((Term) it.next()));
            return Struct.createStructList(clonedElements);
        }
        /*t instanceof WrapStruct*/
        Term[] clonedElements = new Term[((Struct) t).getArity()];
        for (int i = 0; i < clonedElements.length; i++)
            clonedElements[i] = resolveAndCloneStructsImpl(((Struct)t).getArg(i));
        return new Struct((Struct) t, clonedElements);
    }

    public static Term wrapWithID(Term t, int renameVarID){
        if (t instanceof Wrapper)
            return t;
        if (t instanceof Var)
            return new WrapVar((Var) t, renameVarID);
        if (t == Struct.emptyList || t instanceof StructAtom)
            return t;
        if (t instanceof StructList)
            return new WrapStructList((StructList) t, renameVarID);
        if (t instanceof Struct)
            return new WrapStruct((Struct) t, renameVarID);
        return t;
    }



    /**
     * Adds a link.
     * The key of the link is an ID calculated from the Variable (its position in the root parent Struct) and its context
     * The link is stored in three different maps:
     * 1. a map of the link to the Term it is linked to
     * 2. a map of the link to the context number of the Term it is linked to
     * 3. to facilitate garbage collection of unused links,
     * a map linking the context in which the binding takes place to a list of links that are bound at this execution context is also updated
     *
     * @param v
     * @param vCtx
     * @param link
     * @param linkCtx
     * @return true always
     */
    private boolean setLink(Var v, int vCtx, Term link, int linkCtx) /*throws UnificationError */{
        testForWrapperAndThrowOMGException(link);
        if (checkForInternalOccurencesOfVarInTerm(v.getNrInStruct(), vCtx, link, linkCtx))
            return false;
        links.put(v, vCtx, link, linkCtx, currentOwner);
        return true;
    }

    public static Term unWrap(Term l) {
        return l instanceof Wrapper ? ((Wrapper) l).getBasis() : l;
    }

    /**
     * this now only gets in the original terms and the context int as values
     * @param varBeingChecked
     * @param beingCheckedsContext
     * @param t
     * @param tCtx
     */
    private boolean checkForInternalOccurencesOfVarInTerm(int varBeingChecked, int beingCheckedsContext, Term t, int tCtx) {
        if (t instanceof Var){
            if (beingCheckedsContext == tCtx && varBeingChecked == ((Var) t).getNrInStruct())
                return true;
            Term link = links.getTerm((Var)t, tCtx);
            if (link == null)
                return false;
            int linkCtx = links.getCtx((Var)t, tCtx);
            return checkForInternalOccurencesOfVarInTerm(varBeingChecked, beingCheckedsContext, link, linkCtx);
        }
        else if (t instanceof Struct && ((Struct) t).getVarList() != null) {
            for (int i = 0; i < ((Struct) t).getVarList().length; i++) {
                Var v = ((Struct) t).getVarList()[i];
                if (checkForInternalOccurencesOfVarInTerm(varBeingChecked, beingCheckedsContext, v, tCtx))
                    return true;
            }
        }
        return false;
    }

    private static void testForWrapperAndThrowOMGException(Term t) {
        if (t instanceof Wrapper)
            throw new RuntimeException("omg, the BindingsTable has linked to a Wrapper or some Struct containing a Wrapper");
    }

    void collectGarbage(Integer owner) {
        currentOwner = 0;
        links.gcOldLinks(owner);
    }

    void markGarbageCollectionID(int owner) {
        currentOwner = owner;
        links.resetGarbage(owner);
    }

    Iterator getKeysForCp(Integer owner) {
        return links.getKeysForCp(owner);
    }

    public Set getChildrenVars(Term withVars) {
        Set result = new LinkedHashSet();
        if (withVars instanceof Var) {
            result.add(withVars);
        } else if (withVars instanceof StructList) {
            for (Iterator it = structListIterator((Struct) withVars, true); it.hasNext();)
                result.addAll(getChildrenVars((Term) it.next()));
        } else if (withVars instanceof Struct) { //todo, vars are not resolved here.. bug
            for (Iterator it = ((Struct) withVars).children(); it.hasNext();)
                result.addAll(getChildrenVars((Term) it.next()));
        }
        return result;
    }

    /**
     * @param t
     * @return true if t is a Struct that is a list or
     *              if t is a Var that is bound to a Struct that is a list
     * @throws InstantiationError if t is a Var or if the last element of a list t is a Var
     */
    public boolean isList(Term t) throws PrologException {
        t = resolve(t);
        if (t instanceof Var)
            throw new PrologException("instantiation_error");
        if (t == Struct.emptyList)
            return true;
        if (!(t instanceof StructList))
            return false;
        return isList(Struct.getLastChild(((StructList) t)));
    }

    /**
     * @param prolog
     * @param expression
     * @return result of the evaluation as a Number
     * @throws jTrolog.errors.PrologException instantiation error if one of the Variables to be evaluated is not instantiated
     *                         type_error(Evaluable, other) if a part in the evaluable Term is not evaluable
     * @throws Throwable any other exception that might be thrown by parts of the expression to be evaluated
     */
    public Number evalExpression(Prolog prolog, EvaluableTerm expression) throws PrologException, Throwable {
        if (expression instanceof Number)
            return (Number) expression;

        Term child = null;
        try {
            Struct struct = (Struct) expression;
            //chp 9 defines separate behavior for _evaluation_ of float/1 (casting of Int) that should only accept Numbers,
            //as opposed to the execution of float/1 that accepts Term.
            if (struct.hasPredicateIndicator("float", 1)){
                child = resolve(struct.getArg(0));
                Number result = evalExpression(prolog, (EvaluableTerm) child);
                return new Double(result.doubleValue());
            }

            PrimitiveInfo prim = prolog.getPrimitive(struct);
            if (prim == null)
                throw new PrologException("type_error(Evaluable, " + struct.getPredicateIndicator() +")");

            //we have an expression that might be evaluable
            Object[] primitive_args = new Object[struct.getArity()+1];
            primitive_args[0] = this;
            for (int i = 0; i < primitive_args.length-1; i++){
                child = resolve(struct.getArg(i));
                primitive_args[i+1] = evalExpression(prolog, (EvaluableTerm) child);
            }
            return (Number) prim.getMethod().invoke(prim.getSource(), primitive_args);
        } catch (ClassCastException e) {
            if (child instanceof Var)
                throw new PrologException("instantiation_error");
            throw new PrologException("type_error(Evaluable, " + child +")");
        }
    }

    Object[] resolveArgs(Struct unresolved, int ctx) {
        Struct g = (Struct) wrapWithID(unresolved, ctx);
        Object[] primitive_args = new Object[g.getArity()+1];
        primitive_args[0] = this;
        for (int i = 0; i < g.getArity(); i++)
            primitive_args[i+1] = resolve(g.getArg(i));
        return primitive_args;
    }

    int tmpTstFromResolve;
    public Term resolve(Term term) {
        tmpTstFromResolve = -1;
        if (!(term instanceof WrapVar))
            return term;

        Term link = links.getTerm((Var) ((WrapVar) term).getBasis(), ((WrapVar) term).getContext());
        if (link == null)
            return term;

        int linkCtx = links.getCtx((Var) ((WrapVar) term).getBasis(), ((WrapVar) term).getContext());
        term = resolve(wrapWithID(link, linkCtx));
        tmpTstFromResolve = linkCtx;
        return term;
    }

    int secondOutOfFasterResolve;
    public Term resolveFaster(Term term, int ctx) {
        secondOutOfFasterResolve = -1;
        if (!(term instanceof Var))
            return term;

        Term link = links.getTerm((Var) term, ctx);
        if (link == null)
            return term;

        int linkCtx = links.getCtx((Var) term, ctx);
        term = resolveFaster(link, linkCtx);
        secondOutOfFasterResolve = linkCtx;
        return term;
    }

    public Iterator structListIterator(Struct origin, boolean skipLastEmptyList) {
        if (origin == Struct.emptyList)
            return new LinkedList().iterator();
        return new StructListIterator(origin, skipLastEmptyList);
    }

    public static Struct createStructList(Collection complete, BindingsTable unifier) {
        if (complete.isEmpty())
            return Struct.emptyList;
        Struct res = Parser.createNewList(complete.size()-1, unifier.getUniqueExecutionCtxID());
        Iterator toBeLinked = complete.iterator();
        for (Iterator it = unifier.structListIterator(res, true); it.hasNext();) {
            Var child =  (Var) it.next();
            Term link = (Term) toBeLinked.next();
            unifier.unify(child, link);
        }
        return res;
    }

    private class StructListIterator implements Iterator {

        Iterator currentListIterator;
        Term cache;
        boolean includeLastEmptyList;

        StructListIterator(Struct origin, boolean skipLastEmptyList) {
            this.includeLastEmptyList = !skipLastEmptyList;
            currentListIterator = Struct.iterator(((StructList) origin));
        }

        public boolean hasNext() {
            if (cache != null)
                return true;
            if (!currentListIterator.hasNext())
                return false;
            cache = (Term) currentListIterator.next();
            cache = resolve(cache);

            if (currentListIterator.hasNext())
                return true;

            if (cache == Struct.emptyList){
                if (includeLastEmptyList)
                    return true;
                cache = null;
                return false;
            }

            if (!(cache instanceof StructList))
                return true;

            currentListIterator = Struct.iterator(((StructList) cache));
            cache = null;
            return hasNext();
        }

        public Object next() {
            if (cache == null)
                hasNext();
            if (cache == null)
                throw new NoSuchElementException();
            Term temp = cache;
            cache = null;
            return temp;
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }
    }
    //todo override toString();
}