package jTrolog.engine;

import jTrolog.errors.PrologException;
import jTrolog.terms.*;
import jTrolog.terms.Number;

import java.util.*;

/**
 * @author janerist@stud.ntnu.no
 * @author ivar.orstavik@hist.no
 */
class Engine {

    public static final String FALSE = "False".intern();
    public static final String TRUE = "True".intern();
    public static final String TRUE_ALL = "AllTrue".intern();
    public static final String TRUE_NUMBER = "Result_Number".intern();
    public static final String EVAL = "Eval".intern();
    public static final String RULE = "Rule".intern();
    public static final String BACK = "Back".intern();

    private Prolog prolog;
    private Number resultAsNumber = null;

    BindingsTable bt = new BindingsTable();

    private int stackPos = 0;
    private ChoicePoint[] stack = new ChoicePoint[200];

    Engine(Term start, Prolog manager) throws PrologException {
        prolog = manager;
        addToStack(start, 0);
    }

    /**
     * Core of engine. Finite State Machine
     * @param state
     * @return either a Number or a BindingsTable
     */
    Object run(String state) throws Throwable {
        while (true) {
            if (state == EVAL)
                state = eval();
            else if (state == BACK)
                state = back();
            else if (state == RULE)
                state = ruleSelect();
            else if (state == TRUE)
                state = truue();
            else if (state == TRUE_NUMBER)
                return resultAsNumber;
            else if (state == TRUE_ALL)
                return bt;
            else /*state.equals(FALSE)*/
                return null;
        }
    }

    private String truue() throws PrologException {
        for (int i = stackPos-1; i >= 0; i--) {
            ChoicePoint topOfStack = stack[i];
            if(topOfStack.hasRightSideAlternatives){
                topOfStack.hasRightSideAlternatives = false;
                addToStack(topOfStack.struct.getArg(1), topOfStack.getTermCtx());
                stack[stackPos - 1].parentComma = topOfStack;
                return EVAL;
            }
        }
        return TRUE_ALL;
    }

    /**
     * removes comma and cut.
     * Comma => this CP is marked as having a right branch, and the left branch is added to the stack and run first.
     * Cut   => run in separate method
     * Primitive => evaluated
     *      else
     * Rule => rules added, sent to RULE
     */
    private String eval() throws Throwable {
        ChoicePoint topOfStack = stack[stackPos - 1];
        if (topOfStack.struct.hasPredicateIndicator(",", 2)) {
            topOfStack.hasRightSideAlternatives = true;
            addToStack(topOfStack.struct.getArg(0), topOfStack.getTermCtx());
            return EVAL;
        } else if (topOfStack.struct.hasPredicateIndicator("!", 0)) {
            cut();
            return TRUE;
        }
        PrimitiveInfo prim = prolog.getPrimitive(topOfStack.struct);
        if (prim == null){
            topOfStack.setAlternatives(prolog.find(topOfStack.struct.getPredicateIndicator()));
            return topOfStack.hasAlternatives() ? RULE : BACK;
        }
        setupExecCtx(topOfStack);
        Object o = Prolog.evalPrimitive(bt.resolveArgs(topOfStack.struct, topOfStack.getTermCtx()), prim);
        if (o instanceof Number){
            resultAsNumber = (jTrolog.terms.Number) o;
            return TRUE_NUMBER;
        }
        return ((Boolean) o).booleanValue() ? TRUE : BACK;
    }

    private String ruleSelect() throws PrologException {
        ChoicePoint topOfStack = stack[stackPos - 1];
        while (topOfStack.hasAlternatives()) {
            setupExecCtx(topOfStack); //moving this outside the while loop makes things go slower.. why??
            Struct next = (Struct) topOfStack.nextAlternative();
            if (bt.unifyBranch(topOfStack.struct, next.getArg(0), topOfStack.getTermCtx(), stackPos)) {
                addToStack(next.getArg(1), stackPos);
                return EVAL;
            }
            bt.collectGarbage(stackPos);
        }
        return BACK;
    }

    private String back() throws Throwable {
        if (stackPos == 0)
            return FALSE;
        ChoicePoint topOfStack = stack[stackPos - 1];
        bt.collectGarbage(topOfStack.getExecCtx());
        if (topOfStack.hasAlternatives())
            return RULE;
        if (topOfStack.parentComma != null)     //updating the parent comma so that if a new solution is identified in the left side of the comma branch, the backward iteration of the true for loop will try the right side of the comma again
            topOfStack.parentComma.hasRightSideAlternatives = true;
        stackPos--;
        //stack[--stackPos] = null;
        //important line. Here, the CP at the previous stackPos can be deleted.
        //However, constructing new CP objects each time takes time, so by not deleting it,
        //previous CPs can be recycled when they are created (see ChoicePoint.create)
        //Tests not deleting CPs show that it uses up 30% or the total time. But, removing it does not speed up the process.. don't understand why.
        return BACK;
    }

    /**
     * clear all ; and , backtracking alternatives up to and including the first point in the stack that is not , or ; or !.
     */
    private void cut() {
        ChoicePoint commaParent = null;
        for (Iterator it = stackIterator(); it.hasNext();) {
            ChoicePoint next = (ChoicePoint) it.next();
            next.cutAlternatives();

            if (commaParent != null){
                if (next == commaParent)
                    commaParent = next.parentComma;
                //else continue;
            } else {
                if (next.parentComma != null)
                    commaParent = next.parentComma;
                if (!transparentToCut(next))
                    return;
            }
        }
    }

    private boolean transparentToCut(ChoicePoint next) {
        String predIndic = next.struct.getPredicateIndicator();
        if (predIndic.equals("!/0") || predIndic.equals("','/2"))
            return true;
        if (!predIndic.equals("';'/2"))
            return false;
        Term leftArg = next.struct.getArg(0);
        return !(leftArg instanceof Struct && "'->'/2".equals(((Struct) leftArg).getPredicateIndicator()));
    }

    public String toString() {
        String s = "";
        for (Iterator it = stackIterator(); it.hasNext();) {
            ChoicePoint cp = (ChoicePoint) it.next();
            s+= bt.resolveAndCloneStructs(new WrapStruct(cp.struct, cp.getTermCtx()), -1) + "\n";
        }
        return s;
    }

    boolean hasAlternatives() {
        for (Iterator it = stackIterator(); it.hasNext();) {
            if (((ChoicePoint) it.next()).hasAlternatives())
                return true;
        }
        return false;
    }

    private void setupExecCtx(ChoicePoint topOfStack) {
        bt.markGarbageCollectionID(stackPos);
        topOfStack.setExecCtx(stackPos);
    }

    private void addToStack(Term term, int ctx) throws PrologException {
//        if (term instanceof Var)
//            term = new Struct("call", term);
        term = bt.resolveFaster(term, ctx);
        if (bt.secondOutOfFasterResolve != -1)
            ctx = bt.secondOutOfFasterResolve;
        if (term instanceof Number)
            throw new PrologException("type_error(callable, " + term + ")");
        if (term instanceof Var)
            throw new PrologException("instantiation_error");
        ChoicePoint.create(stack, stackPos++,(Struct) term, ctx);
    }

    Iterator reverseIterator() {
        return new Iterator(){
            int pos = 0;
            public boolean hasNext() {
                return pos < stackPos;
            }
            public Object next() {
                return stack[pos++];
            }
            public void remove() {
                throw new RuntimeException("wtf");
            }
        };
    }

    Iterator stackIterator() {
        return new Iterator() {
            int pos = stackPos;
            public boolean hasNext() {
                return pos > 0;
            }
            public Object next() {
                return stack[--pos];
            }
            public void remove() {
                throw new RuntimeException("wtf");
            }
        };
    }
}