/*
 * tuProlog - Copyright (C) 2001-2007 aliCE team at deis.unibo.it
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package jTrolog.lib;

import jTrolog.engine.*;
import jTrolog.terms.Number;
import jTrolog.terms.Float;
import jTrolog.terms.*;
import jTrolog.parser.Parser;
import jTrolog.errors.InvalidLibraryException;
import jTrolog.errors.PrologException;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;

/**
 * This class defines a set of basic built-in
 * predicates for the tuProlog engine
 *
 * Library/Theory dependency: none
 *
 *
 *
 */
public class BasicLibrary extends Library {

    //
    // meta-predicates
    //


    /**
     * sets a new theory provided as a text
     */
    public boolean set_theory_1(BindingsTable bt, Term th) {
        if (!atom_1(bt,th))
            return false;
        try {
            ((Library) this).engine.setTheory(((Struct) th).getName());
            return true;
        } catch (PrologException e) {
            System.err.println("invalid theory:"+e.getMessage());
            return false;
        }
    }

    /**
     *  adds a new theory provided as a text
     */
    public boolean add_theory_1(BindingsTable bt, Term th) throws PrologException {
        if (!atom_1(bt,th))
            return false;
        ((Library) this).engine.addTheory(((Struct) th).getName());
        return true;
    }

    /** gets current theory text */
    public boolean get_theory_1(BindingsTable bt, Term arg) {
        return bt.unify(arg, Struct.createStruct(((Library) this).engine.getTheory()));
    }

    public boolean load_library_2(BindingsTable bt, Struct className, Term libName) throws InvalidLibraryException {
        Library lib = engine.loadLibrary(Parser.removeApices(className.getName()));
        return bt.unify(libName, new StructAtom(lib.getName()));
    }

    public boolean get_operators_list_1(BindingsTable bt, Struct argument) {
        LinkedList result = new LinkedList();
        for (Iterator it = engine.getCurrentOperators(); it.hasNext();)
            result.add((it.next()));
        if (result.isEmpty())
            return bt.unify(argument, Struct.emptyList);
        result.add(Struct.emptyList);
        return bt.unify(argument, BindingsTable.createStructList(result, bt));
    }

    public boolean warning_0(BindingsTable bt) {
        engine.resetWarningList();
        return true;
    }

    public boolean nowarning_0(BindingsTable bt) {
        engine.resetWarningList();
        return true;
    }

    //
    // term type inspection
    //

    public static boolean constant_1(BindingsTable bt, Term t) {
        return atomic_1(bt,t);
    }

    public static boolean number_1(BindingsTable bt, Term t) {
        return t instanceof Number;
    }

    public static boolean integer_1(BindingsTable bt, Term t) {
        return t instanceof Int;
    }

    public static boolean float_1(BindingsTable bt, Term t) {
        return t instanceof Float;
    }

    public static boolean atom_1(BindingsTable bt, Term t) {
        return t instanceof Struct && ((Struct) t).getArity() == 0;
    }

    public static boolean compound_1(BindingsTable bt, Term t) {
        return t instanceof Struct && ((Struct) t).getArity() > 0;
    }

    public static boolean list_1(BindingsTable bt, Term t) {
        try {
            return bt.isList(t);
        } catch (PrologException e) {
            return false;
        }
    }

    public boolean var_1(BindingsTable bt, Term t) {
        return t instanceof Var;
    }

    public boolean nonvar_1(BindingsTable bt, Term t) {
        return !(t instanceof Var);
    }

    public static boolean atomic_1(BindingsTable bt, Term t) {
        if (t instanceof Number)
            return true;
        if (t instanceof Struct)
            return ((Struct) t).getArity() == 0;
        return false;
    }

    public static boolean ground_1(BindingsTable bt, Term t) {
        if (t instanceof Var)
            return false;
        if (t instanceof Struct) {
            for (Iterator it = ((Struct) t).children(); it.hasNext();) {
                Term child = (Term) it.next();
                if (!ground_1(bt, child))
                    return false;
            }
        }
        return true;
    }

    public boolean $arg_3(BindingsTable bt, Int n, Struct term, Term arg) throws PrologException {
        if (!BasicLibrary.compound_1(bt, term))
            throw new PrologException("type_error(compound, " + term + ")");
        if (n.intValue() < 0)
            throw new PrologException("domain_error(not_less_than_zero, " + n + ")");

        if (n.intValue() == 0 || n.intValue() > term.getArity())
            return false;

        Term nthArg = term.getArg(n.intValue() - 1);
        return bt.unify(arg, nthArg);
    }

    public boolean $functor_3(BindingsTable bt, Term term, Term name, Term arity) throws PrologException {
        Int maxArity = (Int) engine.getFlagValue("max_arity");

        if (term instanceof Var) {
            if (name instanceof Var || arity instanceof Var)
                throw new PrologException("instantiation_error");

            if (!BasicLibrary.atomic_1(bt, name))
                throw new PrologException("type_error(atomic, " + name + ")");

            if (!(arity instanceof Int))
                throw new PrologException("type_error(integer, " + arity + ")");

            if (((Int) arity).longValue() > maxArity.intValue())
                throw new PrologException("representation_error(max_arity)");

            if (((Int) arity).intValue() < 0)
                throw new PrologException("domain_error(not_less_than_zero, " + arity + ")");

            if (!BasicLibrary.atom_1(bt, name) && ((Int) arity).intValue() > 0)
                throw new PrologException("type_error(atom, " + name + ")");

            if (BasicLibrary.atomic_1(bt, name) && ((Int) arity).intValue() == 0)
                return bt.unify(term, name);

            Struct newList = Parser.createNewList(((Int) arity).intValue() + 1, bt.getUniqueExecutionCtxID());
            bt.unify(newList.getArg(0), name);
            return $tofromlist_2(bt, term, newList);
        }

        if (BasicLibrary.atomic_1(bt, term))
            return bt.unify(term, name) && bt.unify(arity, new Int(0));

        if (BasicLibrary.compound_1(bt, term)) {
            StructAtom a = new StructAtom(((Struct) term).getName());
            return bt.unify(name, a) && bt.unify(arity, new Int(((Struct) term).getArity()));
        }
        return false;
    }

    public boolean $tofromlist_2(BindingsTable bt, Term arg0, Term arg1) throws PrologException {

        if (arg0 instanceof Number) {
            Struct newList = Parser.createNewList(1, bt.getUniqueExecutionCtxID());
            bt.unify(newList.getArg(0), arg0);
            return bt.unify(newList, arg1);
        }

        if (arg0 instanceof Struct) {
            if (BasicLibrary.compound_1(bt, arg0)) {
                Struct struct = (Struct) arg0;
                WrapStructList medium = Parser.createNewList(struct.getArity() + 1, bt.getUniqueExecutionCtxID());
                Iterator it = bt.structListIterator(medium, false);
                //unify the name of the struct with the first in the medium list
                bt.unify((Term) it.next(), Struct.createStruct(struct.getName()));
                //unify the children of the Struct with the rest of the medium list
                for (Iterator it2 = struct.children(); it2.hasNext();)
                    bt.unify((Var) it.next(), (Term) it2.next());
                //try to unify the medium list generated from the struct with arg1
                return bt.unify(medium, arg1);
            }
        }
        if (arg0 instanceof Var) {
            if (arg1 instanceof Var)
                throw new PrologException("instantiation_error");
            if (!bt.isList(arg1)) //throws instantiation_error
                throw new PrologException("type_error(list, " + arg1 + ")");
            if (arg1 == Struct.emptyList)
                throw new PrologException("domain_error(non_empty_list, " + arg1 + ")");
            Struct list = (Struct) arg1;

            Term head = bt.resolve(list.getArg(0));
            if (head instanceof Var)
                throw new PrologException("instantiation_error");
            if (!BasicLibrary.atom_1(bt, head))
                throw new PrologException("type_error(atom, " + head + ")");

            Term tail = bt.resolve(list.getArg(1));
            if (tail instanceof Struct) {
                LinkedList terms = new LinkedList();
                for (Iterator it = bt.structListIterator((Struct) tail, true); it.hasNext();)
                    terms.add(it.next());
                if (terms.isEmpty())
                    return bt.unify(arg0, head);
                if (terms.size() > ((Int) engine.getFlagValue("max_arity")).intValue())
                    throw new PrologException("representation_error(max_arity)");


                Term res2 = Parser.createStruct(head.toString(), (Term[]) terms.toArray(new Term[0]), bt);
                return bt.unify(arg0, res2);
            } else {
                return bt.unify(arg0, Parser.createStruct(head.toString(), new Term[]{tail}, bt));
            }
        }
        return false;
    }

    //
    // term/espression comparison
    //

    public Term eval_1(BindingsTable bt, EvaluableTerm arg0) throws Throwable {
        return bt.evalExpression(engine, arg0);
    }

    public boolean is_2(BindingsTable bt, Term arg0, EvaluableTerm arg1) throws Throwable {
        return bt.unify(arg0, bt.evalExpression(engine, arg1));
    }

    public boolean expression_equality_2(BindingsTable bt, EvaluableTerm arg0, EvaluableTerm arg1) throws Throwable {
        Number val1 = bt.evalExpression(engine, arg0);
        Number val2 = bt.evalExpression(engine, arg1);
        if (val1 instanceof Float || val2 instanceof Float)
            return Number.compareDoubleValues(val1, val2) == 0;
        return val1.equals(val2);
    }

    public boolean expression_greater_than_2(BindingsTable bt, EvaluableTerm arg0, EvaluableTerm arg1) throws Throwable {
        Number num0 = bt.evalExpression(engine, arg0);
        Number num1 = bt.evalExpression(engine, arg1);
        if (num0 instanceof Float || num1 instanceof jTrolog.terms.Float)
            return Number.compareDoubleValues(num0, num1) >0;
        return num0.intValue() > num1.intValue();
    }

    public boolean expression_less_or_equal_than_2(BindingsTable bt, EvaluableTerm arg0, EvaluableTerm arg1) throws Throwable {
        return ! expression_greater_than_2(bt, arg0, arg1);
    }

    public boolean expression_less_than_2(BindingsTable bt, EvaluableTerm arg0, EvaluableTerm arg1) throws Throwable {
        return !expression_greater_or_equal_than_2(bt, arg0, arg1);
    }

    public boolean expression_greater_or_equal_than_2(BindingsTable bt, EvaluableTerm arg0, EvaluableTerm arg1) throws Throwable {
        return expression_greater_than_2(bt, arg0, arg1) || expression_equality_2(bt, arg0, arg1);
    }

    public static boolean term_equality_2(BindingsTable bt, Term arg0, Term arg1) {
        if (arg0 instanceof Var){
            if (!(arg1 instanceof Var))
                return false;
            Var v = (Var) arg1;
            if (((Var) arg0).isAnonymous() && v.isAnonymous())
                return false;
            return v.equals(arg0);
        }
        if (arg0 instanceof Number){
            if (!(arg1 instanceof Number))
                return false;
            return number_equality_2((Number)arg0, (Number)arg1);
        }

        if (arg0 instanceof Struct){
            if (!(arg1 instanceof Struct))
                return false;

            Struct ts = (Struct) arg1;
            if (((Struct) arg0). getArity() != ts.getArity() || !((Struct) arg0).getName().equals(ts.getName()))
                return false;

            for (int i = 0;i < ((Struct) arg0).getArity();i++) {
                if (!term_equality_2(null, ((Struct) arg0).getArg(i), ts.getArg(i)))
                    return false;
            }
            return true;
        }
        return false;
    }

    public static boolean number_equality_2(Number arg1, Number arg0) {
        if (arg1 instanceof Int && arg0 instanceof Int)
                return arg0.longValue() == arg1.longValue();
        else if (arg1 instanceof Float && arg0 instanceof Float)
            return Number.compareDoubleValues(arg0, arg1) == 0;
        return false;
    }

    public static boolean term_greater_than_2(BindingsTable bt, Term arg0, Term arg1) {
        if (arg0 instanceof Var){
            if (!(arg1 instanceof Var))
                return false;
            return arg0.toString().hashCode() > arg1.toString().hashCode();
        }

        if (arg0 instanceof Struct){
            if (arg1 instanceof Struct) {
                Struct ts = (Struct) arg1;
                int tarity = ts.getArity();
                if (((Struct) arg0).getArity() < tarity)
                    return false;
                if (((Struct) arg0).getArity() == tarity) {
                    if (((Struct) arg0).getName().equals(ts.getName()) ) {
                        for (int i = 0; i < ((Struct) arg0).getArity(); i++) {
                            if (term_greater_than_2(null, ((Struct) arg0).getArg(i), ts.getArg(i)))
                                return true;
                            if (!term_equality_2(null, ((Struct) arg0).getArg(i), ts.getArg(i)))
                                return false;
                        }
                    }
                    if (((Struct) arg0).getName().compareTo(ts.getName()) <= 0)
                        return false;
                }
            }
            return true;
        }
        if (arg0 instanceof Number){
            if (arg1 instanceof Var)
                return true;
            if (arg1 instanceof Struct)
                return false;
            if (!(arg1 instanceof Number))
                return false;

            Number n2 = ((Number) arg1);
            Number n1 = ((Number) arg0);
            if (n1 instanceof Float || n2 instanceof Float)
                return Number.compareDoubleValues(n1, n2) > 0;
            return n1.longValue() > n2.longValue();
        }
        return false; //such as implementation specific NullTerm etc.
    }

    public static boolean term_less_than_2(BindingsTable bt, Term arg0, Term arg1) {
        return !term_greater_than_2(null, arg0, arg1) && !term_equality_2(null, arg0, arg1);
    }


    public Term expression_plus_1(BindingsTable bt, EvaluableTerm arg0) throws Throwable {
        return bt.evalExpression(engine, arg0);
    }

    public Term expression_minus_1(BindingsTable bt, EvaluableTerm arg1) throws Throwable {
        Number val0 = bt.evalExpression(engine, arg1);
        if (val0 instanceof jTrolog.terms.Long)
            return new jTrolog.terms.Long(val0.longValue() * -1);
        if (val0 instanceof Int)
            return new Int(val0.intValue() * -1);
        if (val0 instanceof jTrolog.terms.Double)
            return new jTrolog.terms.Double(val0.doubleValue() * -1);
        if (val0 instanceof jTrolog.terms.Float)
            return new jTrolog.terms.Float(val0.floatValue() * -1);
        return null;
    }

    public Term expression_bitwise_not_1(BindingsTable bt, EvaluableTerm arg0) throws Throwable {
        Number val = bt.evalExpression(engine, arg0);
        if (!(val instanceof Int))
            throw new PrologException("type_error(integer, " + val + ")");
        return new Int(~val.intValue());
    }

    public Term expression_plus_2(BindingsTable bt, EvaluableTerm arg0, EvaluableTerm arg1) throws Throwable {
        Number val0 = bt.evalExpression(engine, arg0);
        Number val1 = bt.evalExpression(engine, arg1);
        if (val0 instanceof Float || val1 instanceof Float)
            return new jTrolog.terms.Double(val0.doubleValue() + val1.doubleValue());
        return Number.getIntegerNumber( val0.longValue() + val1.longValue());
    }

    public Term expression_minus_2(BindingsTable bt, EvaluableTerm arg0,EvaluableTerm arg1) throws Throwable {
        Number val0 = bt.evalExpression(engine, arg0);
        Number val1 = bt.evalExpression(engine, arg1);
        if (val0 instanceof Float || val1 instanceof Float)
            return new jTrolog.terms.Double(val0.doubleValue() - val1.doubleValue());
        return Number.getIntegerNumber(val0.longValue() - val1.longValue());
    }

    public Term expression_multiply_2(BindingsTable bt, EvaluableTerm arg0, EvaluableTerm arg1) throws Throwable {
        Number val0 = bt.evalExpression(engine, arg0);
        Number val1 = bt.evalExpression(engine, arg1);
        if (val0 instanceof Float || val1 instanceof Float)
            return new jTrolog.terms.Double(val0.doubleValue() * val1.doubleValue());
        return Number.getIntegerNumber(val0.longValue() * val1.longValue());
    }

    public Term expression_div_2(BindingsTable bt, EvaluableTerm arg0, EvaluableTerm arg1) throws Throwable {
        Number val0 = bt.evalExpression(engine, arg0);
        Number val1 = bt.evalExpression(engine, arg1);
        if (val0 instanceof Float || val1 instanceof Float)
            return new jTrolog.terms.Double(val0.doubleValue()/val1.doubleValue());
        return Number.getIntegerNumber((long) (val0.doubleValue()/val1.doubleValue()));
    }

    public Term expression_integer_div_2(BindingsTable bt, EvaluableTerm arg0, EvaluableTerm arg1) throws Throwable {
        return Number.getIntegerNumber(bt.evalExpression(engine, arg0).longValue() / bt.evalExpression(engine, arg1).longValue());
    }

    public Term expression_pow_2(BindingsTable bt, EvaluableTerm arg0, EvaluableTerm arg1) throws Throwable {
        double val = Math.pow(bt.evalExpression(engine, arg0).doubleValue(), bt.evalExpression(engine, arg1).doubleValue());
        return new jTrolog.terms.Double(val);
    }

    public Term expression_bitwise_shift_right_2(BindingsTable bt, EvaluableTerm arg0, EvaluableTerm arg1) throws Throwable {
        return new Int(bt.evalExpression(engine, arg0).intValue() >> bt.evalExpression(engine, arg1).intValue());
    }

    public Term expression_bitwise_shift_left_2(BindingsTable bt, EvaluableTerm arg0, EvaluableTerm arg1) throws Throwable {
        return new Int(bt.evalExpression(engine, arg0).intValue() << bt.evalExpression(engine, arg1).intValue());
    }

    public Term expression_bitwise_and_2(BindingsTable bt, EvaluableTerm arg0, EvaluableTerm arg1) throws Throwable {
        return new Int(bt.evalExpression(engine, arg0).intValue() & bt.evalExpression(engine, arg1).intValue());
    }

    public Term expression_bitwise_or_2(BindingsTable bt, EvaluableTerm arg0, EvaluableTerm arg1) throws Throwable {
        return new Int(bt.evalExpression(engine, arg0).intValue() | bt.evalExpression(engine, arg1).intValue());
    }

    //
    // text/atom manipulation predicates
    //

    /**
     * bidirectional text/term conversion.
     */
    public boolean text_term_2(BindingsTable bt, Term arg0, Term arg1) {
        if (!ground_1(bt, arg0))
            return bt.unify(arg0, new StructAtom(arg1.toString()));
        Term result = new Parser(Parser.removeApices(arg0.toString()), engine).nextTerm(false);
        return bt.unify(arg1, result);
    }


    public boolean text_concat_3(BindingsTable bt, Term source1, Term source2, Term dest) {
        if (!atom_1(bt,source1) || !atom_1(bt,source2))
            return false;
        return bt.unify(dest, new StructAtom(((Struct) source1).getName() + ((Struct) source2).getName()));
    }

    public boolean num_atom_2(BindingsTable bt, Term arg0,Term arg1) {
        if (arg1 instanceof Var) {
            if (arg0 instanceof Number)
                return bt.unify(arg1, new StructAtom(arg0.toString()));
            return false;
        }
        if (!atom_1(bt,arg1))
            return false;
        Number num = Parser.parseNumber(Parser.removeApices(((Struct) arg1).getName()).trim());
        return bt.unify(arg0, num);
    }

    public String getTheory(){
        return
        //
        // operators defined by the BasicLibrary theory
        //
        "':-'(op( 1200, fx,   ':-')). \n"+
        ":- op( 1200, xfx,  ':-'). \n"+
        ":- op( 1200, fx,   '?-'). \n"+
        ":- op( 1100, xfy,  ';'). \n"+
        ":- op( 1050, xfy,  '->'). \n"+
        ":- op( 1000, xfy,  ','). \n"+
        ":- op(  900, fy,   '\\+'). \n"+
        ":- op(  900, fy,   'not'). \n"+
        //
        ":- op(  700, xfx,  '='). \n"+
        ":- op(  700, xfx,  '\\='). \n"+
        ":- op(  700, xfx,  '=='). \n"+
        ":- op(  700, xfx,  '\\=='). \n"+
        //
        ":- op(  700, xfx,  '@>'). \n"+
        ":- op(  700, xfx,  '@<'). \n"+
        ":- op(  700, xfx,  '@=<'). \n"+
        ":- op(  700, xfx,  '@>='). \n"+
        ":- op(  700, xfx,  '=:='). \n"+
        ":- op(  700, xfx,  '=\\='). \n"+
        ":- op(  700, xfx,  '>'). \n"+
        ":- op(  700, xfx,  '<'). \n"+
        ":- op(  700, xfx,  '=<'). \n"+
        ":- op(  700, xfx,  '>='). \n"+
        //
        ":- op(  700, xfx,  'is'). \n"+
        ":- op(  700, xfx,  '=..'). \n"+
        ":- op(  500, yfx,  '+'). \n"+
        ":- op(  500, yfx,  '-'). \n"+
        ":- op(  500, yfx,  '/\\'). \n"+
        ":- op(  500, yfx,  '\\/'). \n"+
        ":- op(  400, yfx,  '*'). \n"+
        ":- op(  400, yfx,  '/'). \n"+
        ":- op(  400, yfx,  '//'). \n"+
        ":- op(  400, yfx,  '>>'). \n"+
        ":- op(  400, yfx,  '<<'). \n"+
        ":- op(  400, yfx,  'rem'). \n" +
        ":- op(  400, yfx,  'mod'). \n" +
        ":- op(  200, xfx,  '**'). \n"+
        ":- op(  200, xfy,  '^'). \n"+
        ":- op(  200, fy,   '\\'). \n"+
        ":- op(  200, fy,   '-'). \n"+
        //
        // flag management
        //
        "current_prolog_flag(Name,Value) :- get_prolog_flag(Name,Value),!.\n"+
        "current_prolog_flag(Name,Value) :- flag_list(L), member(flag(Name,Value),L).\n"+
        //
        // expression/term comparison
        //
        "'=\\='(X,Y):- not expression_equality(X,Y). \n"+
        "'\\=='(X,Y):- not term_equality(X,Y).\n"+
        "'@>='(X,Y):- not term_less_than(X,Y).\n"+
        "'@=<'(X,Y):- not term_greater_than(X,Y).\n"+
        //
        // meta-predicates
        //
        "clause(H,B) :- (nonvar(H) ; '$instantiation_error'), (not(number(H)) ; '$type_error'(callable, H)), (not(number(B)) ; '$type_error'(callable, B)), '$find'(H,L), member((':-'(H,B)),L). \n                                            " +
        "current_predicate(PI) :- PI = Name/Arity, " +
                                "'$all_dynamic_predicate_indicators'(Iterator), " +
                                "has_next(Iterator), " +
                                "'$current_pred_impl'(Name, Arity, Iterator). " +

        "has_next(Iterator) :- not('$has_next'(Iterator)), !, fail. " +
        "has_next(Iterator) :- true. " +
        "has_next(Iterator) :- has_next(Iterator). " +

        "copy_term(T,V):-'$copy'(T,V). \n                                                                        " +
        //
//        "call(P):- var(P), '$instantiation_error'. \n"+
//        "call(P):- number(P), '$type_error'(callable, P). \n"+
        "call(P):- P. \n"+
        "'\\+'(P):- call(P),!,fail.\n                                                                            "+
        "'\\+'(_).\n                                                                                             "+
        "C -> T ; B :- !, or((C, !, T), B). \n" +
        "C -> T :- call(C), !, call(T). \n" +
        "or(A, B) :- A. \n" +
        "or(A, B) :- B. \n" +
        "A ; B :- A =.. ['->', C, T], !, (C, !, T ; B). \n                                                              "+
        "A ; B :- A. \n                                                                                          "+
        "A ; B :- B. \n                                                                                          "+
        "unify_with_occurs_check(X,Y):-X=Y.\n                                                                     "+
        "current_op(Pri,Type,Name):-get_operators_list(L),member(op(Pri,Type,Name),L).\n                          "+
        "once(X) :- myonce(X).\n                                                                                  "+
        "myonce(X):-X,!.\n                                                                                        "+
        "repeat. \n                                                                                              " +
        "repeat        :- repeat. \n                                                                             " +
        "not(G)        :- G,!,fail. \n                                                                     " +
        "not(_). \n                                                                                              " +
        //
        // All solutions predicates
        //
        "findall(Template, Goal, Instances) :- \n" +
//            "L = [], \n" +
            "findall_impl(Template, Goal, L), \n" +
            "Instances = L. \n" +
        "findall_impl(Template, Goal, L) :- \n" +
            "call(Goal), \n" +
            "'$copy'(Template, CL), \n" +
            "'$append_in'(CL), \n" +
            "fail. \n" +
        "findall_impl(_, _, Instances) :- '$append_out'(Instances). \n" +

        "bagof(Template, Goal, Instances) :- \n" +
            "iterated_goal_term(Goal, G), \n" +
            "free_variables_set(Goal, Template, Set), \n" +
            "Witness =.. [witness | Set], \n" +
            "findall(Witness + Template, G, S), \n" +
            "'$bagof_impl_a'(Witness, S, Instances). \n" +

        "'$bagof_impl_a'(_, [], _) :- !, fail. \n" +
        "'$bagof_impl_a'(WitnessIn, BigSet, Instances) :- " +
                "'$stripBagList'(WitnessIn, BigSet, Matches, RemainderSet, Variant), " +
                "'$bagof_impl_b'(WitnessIn, Instances, RemainderSet, Variant, Matches). " +

        "'$bagof_impl_b'(WitnessIn, Instances, _, Variant, Matches) :- WitnessIn = Variant, Instances = Matches. \n" +  //first success?
        "'$bagof_impl_b'(WitnessIn, Instances, RemainderSet, _, _) :- '$bagof_impl_a'(WitnessIn, RemainderSet, Instances). \n" +  //get the next from remainding bag

        "setof(Template, Goal, Instances) :- \n" +
            "bagof(Template, Goal, List), \n" +
            "quicksort(List, '@<', OrderedList), \n" +
            "no_duplicates(OrderedList, Instances). \n" +

        "variable_set(Term, List) :- '$variable_set'(Term, List). \n" +

        "free_variables_set(Term, WithRespectTo, Set) :- \n" +
            "variable_set(Term, VS), \n" +
            "variable_set(WithRespectTo, VS1), \n" +
            "existential_variables_set(Term, EVS1), \n" +
            "'$list_diff'(VS, VS1, T), \n" +
            "'$list_diff'(T, EVS1, T2), \n" +
            "Set =T2. \n" +

        "existential_variables_set(Term, []) :- var(Term), !. \n" +
        "existential_variables_set(Term, []) :- atomic(Term), !. \n" +
        "existential_variables_set(V ^ G, EVS) :- \n" +
            "variable_set(V, VS), existential_variables_set(G, ExistentialVars), !, append(VS, ExistentialVars, EVS). \n" +
        "existential_variables_set(_, []). \n" +

        "iterated_goal_term(Var, _) :- var(Var), '$instantiation_error'. \n" +
        "iterated_goal_term(Number, _) :- number(Number), '$type_error'(callable, Number). \n"+
        "iterated_goal_term(_ ^ SubGoal, Goal) :- iterated_goal_term(SubGoal, Goal). \n" +
        "iterated_goal_term(G, G). \n" +

        "no_duplicates([], []). \n" +
        "no_duplicates([H | T], L) :- is_member(H, T), !, no_duplicates(T, L). \n" +
        "no_duplicates([H | T], [H | L]) :- no_duplicates(T, L). \n" +
        "is_member(E, [H | _]) :- E == H, !. \n" +
        "is_member(E, [_ | T]) :- is_member(E, T). \n" +
        //
        // theory management predicates
        //
        "assert(C) :- assertz(C). \n" +
        "abolish(P/A) :- abolish(P,A). \n" +
        "abolish(C) :- '$type_error'(predicate_indicator, C). \n" +
        "retract(Rule) :- Rule = ':-'(Head, Body), !, clause(Head, Body), '$retract'(Rule). \n" +
        "retract(Fact) :- clause(Fact, true), '$retract'(Fact). \n" +

        "retractall(Head) :- findall(Head, clause(Head, _), L), '$retract_clause_list'(L), !. \n" +
        "'$retract_clause_list'([]). \n" +
        "'$retract_clause_list'([E | T]) :- !, '$retract'(E), '$retract_clause_list'(T). \n" +
        //
        // auxiliary predicates
        //
        "member(E,[E|_]). \n                                                                                     " +
        "member(E,[_|L]):- member(E,L). \n                                                                       " +
        "length(L, S) :- number(S), S >= 0, !, lengthN(L, S), !. \n" +
        "length(L, S) :- var(S), lengthX(L, S). \n" +
        "lengthN([],0). \n" +
        "lengthN(_, N) :- nonvar(N), N < 0, !, fail. \n" +
        "lengthN([_|L], N) :- lengthN(L,M), N is M + 1. \n" +
        "lengthX([],0). \n" +
        "lengthX([_|L], N) :- lengthX(L,M), N is M + 1. \n" +
        "append([],L2,L2). \n                                                                                    " +
        "append([E|T1],L2,[E|T2]):- append(T1,L2,T2). \n                                                         " +
        "reverse(L1,L2):- reverse0(L1,[],L2). \n                                                                 " +
        "reverse0([],Acc,Acc). \n                                                                                " +
        "reverse0([H|T],Acc,Y):- reverse0(T,[H|Acc],Y). \n                                                       " +
        "delete(E,[],[]). \n                                                                                     " +
        "delete2(E,[E|T],L):- !,delete(E,T,L). \n                                                                 " +
        "delete(E,[H|T],[H|L]):- delete2(E,T,L). \n                                                               " +
        "element(1,[E|L],E):- !. \n                                                                              " +
        "element(N,_,_):- N < 0, !, fail. \n                                                                     " +
        "element(N,[_|L],E):- M is N - 1,element(M,L,E). \n                                                      " +

        "quicksort([],Pred,[]).                             \n" +
        "quicksort([X|Tail],Pred,Sorted):-                  \n"+
        "   split(X,Tail,Pred,Small,Big),                   \n"+
        "   quicksort(Small,Pred,SortedSmall),              \n"+
        "   quicksort(Big,Pred,SortedBig),                  \n"+
        "   append(SortedSmall,[X|SortedBig],Sorted).       \n"+
        "split(_,[],_,[],[]).                               \n"+
        "split(X,[Y|Tail],Pred,Small,[Y|Big]):-             \n"+
        "   Predicate =..[Pred,X,Y],                        \n"+
        "   call(Predicate),!,                              \n"+
        "   split(X,Tail,Pred,Small,Big).                   \n"+
        "split(X,[Y|Tail],Pred,[Y|Small],Big):-             \n"+
        "   split(X,Tail,Pred,Small,Big).                   \n";
    }

    // Internal Java predicates which are part of the bagof/3 and setof/3 algorithm

    public boolean $has_next_1(BindingsTable bt, IteratorAsTerm iterator) {
        return iterator.hasNext();
    }

    public boolean $all_dynamic_predicate_indicators_1(BindingsTable bt, Term iterator) throws PrologException {
        if (iterator instanceof Var){
            Term iteratorTerm = new IteratorAsTerm(engine.dynamicPredicateIndicators());
            return bt.unify(iterator, iteratorTerm);
        }
        if (iterator instanceof IteratorAsTerm)
            return ((IteratorAsTerm) iterator).hasNext();
        throw new PrologException("$all_dynamic_predicate_indicators has a bug. contact ivar.");
    }

    public boolean $current_pred_impl_3(BindingsTable bt, Term name, Term arity, IteratorAsTerm allDynamicPIs) {
        String nextDynamicPI = (String) allDynamicPIs.next();
        String[] pred = nextDynamicPI.split("/");
        if (!bt.unify(name, new StructAtom(pred[0])))
            return false;
        return bt.unify(arity, new Int(pred[1]));
    }

    public boolean $variable_set_2(BindingsTable bt, Term withVars, Term varList) {
        Set newVarList = bt.getChildrenVars(withVars);
        newVarList.add(Struct.emptyList);
        return bt.unify(BindingsTable.createStructList(newVarList, bt), varList);
    }

    private LinkedList appendStorage;
    // $append
    public boolean $append_in_1(BindingsTable bt, Term unit) {
        if (appendStorage == null)
            appendStorage = new LinkedList();
        appendStorage.add(unit);
        return true;
    }

    public boolean $append_out_1(BindingsTable bt, Term list) {
        if (appendStorage == null)
            return bt.unify(list, Struct.emptyList);
        appendStorage.add(Struct.emptyList);
        boolean res = bt.unify(list, BindingsTable.createStructList(appendStorage, bt));
        appendStorage = null;
        return res;
    }

    public boolean $stripBagList_5(BindingsTable bt, Struct witnessIn, Struct a_and_bSet, Term aSet, Term bSet, Term variant) {
        LinkedList results = new LinkedList();
        LinkedList remains = new LinkedList();

        int wrapID = witnessIn instanceof WrapStruct ? ((WrapStruct) witnessIn).context : bt.getUniqueExecutionCtxID();
        witnessIn = (Struct) bt.resolveAndCloneStructs(witnessIn, wrapID);

        //1a. get a variant of the first witness+template from the inputList
        Iterator it = bt.structListIterator(a_and_bSet, true);
        Struct w_t = (Struct) it.next();

        //1b. the witness branch of the first result will be used to match other results with identical witnesses
        Struct witness1 = (Struct) variant(witnessIn, w_t.getArg(0));
        bt.unify(w_t.getArg(0), witness1);


        //1c. unify first witness with Variant output and add to result
        bt.unify(variant, witness1);
        results.add(BindingsTable.unWrap(w_t.getArg(1)));

        //2. iterate the rest of the input list and
        //   put all matching witnesses from bigset into result and the rest in remainder
        while ( it.hasNext()) {
            w_t = (Struct) it.next();
            Struct witness2 = (Struct) variant(witness1, w_t.getArg(0));
            if (term_equality_2(null, witness1, witness2)){
                bt.unify(w_t.getArg(0), witness1);
                results.add(BindingsTable.unWrap(w_t.getArg(1)));
            } else {
                remains.add(BindingsTable.unWrap(w_t));
            }
        }

        //3. unify the results lists and remains list with aSet and bSet
        if (remains.isEmpty()) {
            bt.unify(bSet, Struct.emptyList);
        } else {
            remains.add(Struct.emptyList);
            bt.unify(bSet, BindingsTable.wrapWithID(Struct.createStructList(remains), wrapID));
        }
        results.add(Struct.emptyList);
        return bt.unify(aSet, BindingsTable.wrapWithID(Struct.createStructList(results), wrapID));
    }

    //todo not finished.. Just a hack to see if things roughly work
    //todo bug in the bag_of predicate..
    //todo I think the fault lies with variant..
    //todo I think variant constructs a Struct with Variables from various contexts..
    private static Term variant(Term former, Term latter) {
        if (BasicLibrary.atomic_1(null, latter))
            return latter;
        if (latter instanceof Var){
            if (former instanceof Var)
                return former;
            throw new RuntimeException("h");
        }
        if (latter instanceof Struct) {
            if (former instanceof Var)
                return latter;
            if (!(former instanceof Struct))
                throw new RuntimeException("h2");
            if (!((Struct) former).hasPredicateIndicator(((Struct) latter).getName(), ((Struct) latter).getArity()))
                throw new RuntimeException("h3");
            Term[] newChildren = new Term[((Struct) latter).getArity()];
            Iterator one = ((Struct) former).children();
            Iterator two = ((Struct) latter).children();
            for (int i = 0; one.hasNext();i++)
                newChildren[i] = variant(BindingsTable.unWrap((Term) one.next()), BindingsTable.unWrap((Term) two.next())); //todo this is just a hack. I don't think it works at all, this needs to be fixed
            return BindingsTable.wrapWithID(/*new WrapStruct(*/new Struct(((Struct) latter).getName(), newChildren), ((WrapStruct) former).context); //todo bug in context setting here..
        }
        throw new RuntimeException("h4");
    }

    public boolean $list_diff_3(BindingsTable bt, Struct main, Struct retract, Var diff) {
        if (retract == Struct.emptyList)
            return bt.unify(diff, main);
        LinkedList resultList = new LinkedList();
        outer : for (Iterator it = bt.structListIterator(main, true); it.hasNext();) {
            Term child = (Term) it.next();
            for (Iterator it2 = bt.structListIterator(retract, true); it2.hasNext();) {
                Term child2 = (Term) it2.next();
                if (BasicLibrary.term_equality_2(null,child, child2))
                    continue outer;
            }
            resultList.add(child);
        }
        if (resultList.isEmpty())
            return bt.unify(diff, Struct.emptyList);
        resultList.add(Struct.emptyList);
        return bt.unify(diff, BindingsTable.createStructList(resultList, bt));
    }


    /**
     * Defines a map for synonyms for primitives.
     * String primitive name -> String[]{synonym name, another synonym name, ..}.
     */
    public String[] getSynonym(String primitive) {
        if (primitive.equals("expression_plus"))
            return new String[]{"+"};
        if (primitive.equals("expression_minus"))
            return new String[]{"-"};
        if (primitive.equals("expression_multiply"))
            return new String[]{"*"};
        if (primitive.equals("expression_div"))
            return new String[]{"/"};
        if (primitive.equals("expression_pow"))
            return new String[]{"**"};
        if (primitive.equals("expression_bitwise_shift_right"))
            return new String[]{">>"};
        if (primitive.equals("expression_bitwise_shift_left"))
            return new String[]{"<<"};
        if (primitive.equals("expression_bitwise_and"))
            return new String[]{"/\\"};
        if (primitive.equals("expression_bitwise_or"))
            return new String[]{"\\/"};
        if (primitive.equals("expression_integer_div"))
            return new String[]{"//"};
        if (primitive.equals("expression_bitwise_not"))
            return new String[]{"\\"};
        if (primitive.equals("$functor"))
            return new String[]{"functor"};
        if (primitive.equals("$arg"))
            return new String[]{"arg"};
        if (primitive.equals("$tofromlist"))
            return new String[]{"=.."};
        if (primitive.equals("expression_equality"))
            return new String[]{"=:="};
        if (primitive.equals("expression_greater_than"))
            return new String[]{">"};
        if (primitive.equals("expression_less_than"))
            return new String[]{"<"};
        if (primitive.equals("expression_greater_or_equal_than"))
            return new String[]{">="};
        if (primitive.equals("expression_less_or_equal_than"))
            return new String[]{"=<"};
        if (primitive.equals("term_equality"))
            return new String[]{"=="};
        if (primitive.equals("term_greater_than"))
            return new String[]{"@>"};
        if (primitive.equals("term_less_than"))
            return new String[]{"@<"};
        return null;
    }

}