/*
 * tuProlog - Copyright (C) 2001-2007 aliCE team at deis.unibo.it
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package jTrolog.lib;

import jTrolog.errors.InvalidLibraryException;
import jTrolog.errors.PrologException;
import jTrolog.parser.Parser;
import jTrolog.engine.*;
import jTrolog.terms.*;
import jTrolog.terms.Number;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;


/**
 * Library of built-in predicates
 *
 * @author Alex Benini
 * @author janerist@stud.ntnu.no
 * @author ivar.orstavik@hist.no
 */
public class BuiltIn extends Library {

    private Prolog engine;

    public BuiltIn(Prolog mediator) {
        engine = mediator;
    }

    /**
     * Defines a map for synonyms for primitives.
     * String primitive name -> String[]{synonym name, another synonym name, ..}.
     */
    public String[] getSynonym(String primitive) {
        if (primitive.equals("cut"))
            return new String[]{"!"};
        if (primitive.equals("unify"))
            return new String[]{"="};
        if (primitive.equals("deunify"))
            return new String[]{"\\="};
        if (primitive.equals("comma"))
            return new String[]{","};
        if (primitive.equals("solve"))
            return new String[]{"initialization"};
        if (primitive.equals("$load_library"))
            return new String[]{"load_library"};
        return null;
    }

    /**
     * Primitives
     */
    public boolean comma_2(BindingsTable bt, Term arg0, Term arg1) {
        throw new RuntimeException("','/2 method is hardcoded in the Engine, do not use Buitlin.comma_2!");
    }

    public static boolean cut_0(BindingsTable bt) {
        throw new RuntimeException("!/0 method is hardcoded in engine, do not use Buitlin.cut_0!");
    }

    public static boolean fail_0(BindingsTable bt) {
        return false;
    }

    public static boolean true_0(BindingsTable bt) {
        return true;
    }

    public static boolean halt_0(BindingsTable bt) throws PrologException {
        throw new PrologException("halt()");
    }

    public boolean halt_1(BindingsTable bt, jTrolog.terms.Number arg0) throws PrologException {
        throw new PrologException("halt("+arg0.intValue() +")");
    }

    public boolean $debug_1(BindingsTable bt, Term arg0) {
        System.out.println("debug=== " + arg0);
        return true;
    }

    public boolean $debug_3(BindingsTable bt, Term arg0, Term arg2, Term arg3) {
        System.out.println("debug0=== " + arg0);
        System.out.println("debug1=== " + arg2);
        System.out.println("debug2=== " + arg3);
        return true;
    }

    public boolean asserta_1(BindingsTable bt, Term arg0) throws PrologException {
        if (arg0 instanceof Var)
            throw new PrologException("instantiation_error");

        if (!(arg0 instanceof Struct))
            throw new PrologException("type_error(callable, " + arg0 + ")");

        Struct struct = makeClause((Struct) arg0);
        Struct cl = (Struct) bt.resolveAndCloneStructs(struct, 0);
        Term head = cl.getArg(0);
        Term body = cl.getArg(1);

        if (head instanceof Var)
            throw new PrologException("instantiation_error");
        if (!(head instanceof Struct))
            throw new PrologException("type_error(callable, " + head + ")");
        if (!(body instanceof Struct)) // TODO: can body be Var?
            throw new PrologException("type_error(callable, " + body + ")");

        engine.assertA(cl);
        return true;
    }

    public boolean assertz_1(BindingsTable bt, Term arg0) throws PrologException {
        if (arg0 instanceof Var)
            throw new PrologException("instantiation_error");

        if (!(arg0 instanceof Struct))
            throw new PrologException("type_error(callable, " + arg0 + ")");

        Struct struct = makeClause((Struct) arg0);
        Struct cl = (Struct) bt.resolveAndCloneStructs(struct, 0);
        Term head = cl.getArg(0);
        Term body = cl.getArg(1);

        if (head instanceof Var)
            throw new PrologException("instantiation_error");
        if (!(head instanceof Struct))
            throw new PrologException("type_error(callable, " + head + ")");
        if (!(body instanceof Struct)) // TODO: can body be Var?
            throw new PrologException("type_error(callable, " + body + ")");

        engine.assertZ(cl);
        return true;
    }

    private static Struct makeClause(Struct struct) {
        if (!Struct.isClause(struct))
            struct = (Struct) BindingsTable.wrapWithID(new Struct(":-", ((WrapStruct) struct).getBasis(), Term.TRUE), ((WrapStruct) struct).getContext());
        return struct;
    }

    public boolean $retract_1(BindingsTable bt, Struct arg0) throws PrologException {
        arg0 = makeClause(arg0);
        Term head = arg0.getArg(0);
        if (head instanceof Var)
            throw new PrologException("instantiation_error");
        if (!(head instanceof Struct))
            throw new PrologException("type_error(callable, " + head + ")");

        Struct sarg0 = (Struct) bt.resolveAndCloneStructs(arg0, bt.getUniqueExecutionCtxID());
        Struct retractedClause = engine.retract(sarg0);
        // if clause to retract found -> retract + true
        if (retractedClause != null) {
            bt.unify(retractedClause, arg0);
            return true;
        }
        return false;
    }

    /**
     * see 8.9.4 p.81 in ISO standard
     *
     * @param name
     * @return false if error occured, true otherwise, regardless of wether or not any predicates were abolished (always true)
     * @throws PrologException
     */
    public boolean abolish_2(BindingsTable bt, Struct name, Int arity) throws PrologException {
        if (!BasicLibrary.atom_1(bt, name))
            throw new PrologException("type_error(atom, " + name + ")");
        engine.abolish(name.toString() + "/" + arity);
        return true;

    }

    /*
      * loads a engine library, given its java
      * class name
      */
    public boolean load_library_1(BindingsTable bt, Struct arg0) throws InvalidLibraryException {
        if (!BasicLibrary.atom_1(bt, arg0))
            return false;
        engine.loadLibrary(arg0.getName());
        return true;
    }

    /**
     * unloads a engine library, given its java class name
     */
    public boolean unload_library_1(BindingsTable bt, Struct arg0) throws InvalidLibraryException {
        if (!BasicLibrary.atom_1(bt, arg0))
            return false;
        engine.unloadLibrary(arg0.getName());
        return true;
    }

    /*
      * get flag list:  flag_list(-List)
      */
    public boolean flag_list_1(BindingsTable bt, Term arg0) {
        return bt.unify(arg0, BindingsTable.wrapWithID(engine.getPrologFlagList(), bt.getUniqueExecutionCtxID()));
    }


    public boolean unify_2(BindingsTable bt, Term arg0, Term arg1) {
        return bt.unify(arg0, arg1);
    }


    // \=
    public boolean deunify_2(BindingsTable bt, Term arg0, Term arg1) {
        return !bt.unify(arg0, arg1);
    }


    //unifies a new StructList of the given length with Any as elements with out
    public boolean newlist_2(BindingsTable bt, Int length, Var out) throws PrologException {
        int lengthInt = length.intValue();
        if (lengthInt < 0)
            throw new PrologException("domain_error(not_less_than_zero, " + length + ")");
        if (lengthInt == 0)
            return bt.unify(out, Struct.emptyList);
        return bt.unify(out, Parser.createNewList(lengthInt, bt.getUniqueExecutionCtxID()));
    }

    // $copy
    public static int cloneNumber = -1;
    private int varNumberCounter;

    public boolean $copy_2(BindingsTable bt, Term arg0, Term arg1) {
        varNumberCounter = 1;
        Term copy = copyAndSetNewWrapperIDOnBothStructsAndVars(arg0, bt);
        copy = BindingsTable.wrapWithID(copy, bt.getUniqueExecutionCtxID());
        return bt.unify(arg1, copy);
    }

    private Term copyAndSetNewWrapperIDOnBothStructsAndVars(Term original, BindingsTable bt) {
        if (!(original instanceof Wrapper))
            return original;
        if (original instanceof WrapVar) {
            Term link = bt.resolve(original);
            if (link != original)
                return copyAndSetNewWrapperIDOnBothStructsAndVars(link, bt);
            return new Var(((WrapVar) original).getBasis().toString() + "_" + ((WrapVar) original).getContext(), varNumberCounter++);
        }
        if (original instanceof WrapStructList) {
            LinkedList clonedElements = new LinkedList();
            for (Iterator it = bt.structListIterator((Struct) original, false); it.hasNext();)
                clonedElements.add(copyAndSetNewWrapperIDOnBothStructsAndVars((Term) it.next(), bt));
            return Struct.createStructList(clonedElements);
        }
        /*t instanceof WrapStruct*/
        Term[] clonedElements = new Term[((WrapStruct) original).getArity()];
        for (int i = 0; i < clonedElements.length; i++)
            clonedElements[i] = copyAndSetNewWrapperIDOnBothStructsAndVars(((WrapStruct) original).getArg(i), bt);
        return new Struct((Struct) ((WrapStruct) original).getBasis(), clonedElements);
    }

    // $find
    // look for clauses whose head unifies whith arg0 and unify the list of them with arg1
    public boolean $find_2(BindingsTable bt, Struct arg0, Term arg1) throws PrologException, CloneNotSupportedException {
        String key = arg0.getPredicateIndicator();
        if (engine.hasPrimitive(key))
            throw new PrologException("permission_error(access, static_procedure, " + key + ")");
        LinkedList res = new LinkedList();
        for (Iterator it = engine.find(key).iterator(); it.hasNext();)
            res.add((Struct) it.next());
        res.add(Struct.emptyList);
        return bt.unify(arg1, BindingsTable.wrapWithID(Struct.createStructList(res), bt.getUniqueExecutionCtxID()));
    }

    // set_prolog_flag(+Name,@Value)
    public boolean set_prolog_flag_2(BindingsTable bt, Term arg0, Term arg1) {
        if (!BasicLibrary.atom_1(bt, arg0) && !(arg0 instanceof Struct) || !BasicLibrary.ground_1(bt, arg1))
            return false;

        String name = arg0.toString();
        Flag flag = engine.getFlag(name);//Flag) it.next();
        if (flag != null) {
            for (java.util.Iterator it2 = bt.structListIterator(flag.getValueList(), true); it2.hasNext();) {
                Term t = (Term) it2.next();
                if (Prolog.match(arg1, t)) {
                    flag.setValue(arg1);
                    return true;
                }
            }
        }
        return false;
    }


    // get_prolog_flag(@Name,?Value)
    public boolean get_prolog_flag_2(BindingsTable bt, Term arg0, Term arg1) {
        if (!BasicLibrary.atom_1(bt, arg0) && !(arg0 instanceof Struct))
            return false;

        String name = arg0.toString();
        Term value = engine.getFlagValue(name);
        if (value == null)
            return false;
        return bt.unify(value, arg1);
    }

    /*
      * DIRECTIVES
      */

    /**
     * op(+Precedence, +Type, +Name) defines a new operator
     * if precedence not in 0..1200 -> delete currently present op
     */
    public void op_3(BindingsTable bt, Number arg0, StructAtom arg1, StructAtom arg2) {
        String st = arg1.getName();
        if (st.matches("[xy]?f[xy]?"))      //a tad simplified
            engine.opNew(arg2.toString(), st, arg0.intValue());
    }


    public void flag_4(BindingsTable bt, Term flagName, Struct flagSet, Term flagDefault, Term flagModifiable) {
        if (flagModifiable.equals(Term.TRUE) || flagModifiable.equals(Term.FALSE))
            engine.defineFlag(flagName.toString(), flagSet, flagDefault, flagModifiable.equals(Term.TRUE));
    }

    //todo test that this works properly.
    //todo Can alternatively use a temporary list to be run after the rest of the theories in consults have been added.
    public void solve_1(BindingsTable bt, Struct goal) throws Throwable {
        engine.solve(bt.resolveAndCloneStructs(goal, bt.getUniqueExecutionCtxID()).toString());
    }

    public void $load_library_1(BindingsTable bt, Term lib) throws InvalidLibraryException {
        if (BasicLibrary.atom_1(bt, lib))
            engine.loadLibrary(((Struct) lib).getName());
    }

    public void consult_1(BindingsTable bt, Term theory) throws FileNotFoundException, PrologException, IOException {
        FileInputStream is;
        try {
            is = new FileInputStream(Parser.removeApices(theory.toString()));
        } catch (FileNotFoundException e) {
            File dir = new File(System.getProperty("user.dir"));
            String filename = dir + File.separator + Parser.removeApices(theory.toString());
            is = new FileInputStream(filename);
        }
        engine.addTheory(IOLibrary.readStream(is));
    }

    public boolean $instantiation_error_0(BindingsTable bt) throws PrologException {
        throw new PrologException("instantiation_error");
    }

    public boolean $type_error_2(BindingsTable bt, Term typeName, Term term) throws PrologException {
        throw new PrologException("type_error(" + typeName + ", " + term + ")");
    }

    public boolean $representation_error_1(BindingsTable bt, Term maxArity) throws PrologException {
        throw new PrologException("representation_error(" + maxArity + ")");
    }

    public boolean $domain_error_zero_1(BindingsTable bt, Term term) throws Exception {
        throw new PrologException("domain_error(not_less_than_zero, " + term + ")");
    }
}