package jTrolog.lib;

import jTrolog.engine.BindingsTable;
import jTrolog.parser.Parser;
import jTrolog.errors.InvalidTermException;
import jTrolog.errors.PrologException;
import jTrolog.terms.*;

import java.util.Iterator;
import java.util.LinkedList;

/**
 * Library for managing DCGs.
 *
 * 2p DCG tutorial:
 *
 * 1. add the DCG library to you'r Prolog engine
 *      Prolog p = new Prolog("jTrolog.lib.DCGLibrary");
 *
 * 2. add the DCG theory as dynamic theories, not as part of a static library.
 *      p.addTheory(new Theory("n --> [N], {noun(N)}. noun(cat). "));
 *
 * 3. prepare your DCG clauses:
 *      p.solve("dcg_preprocess.");
 *
 * 4. use your DCG clauses:
 *      p.solve("n([cat], B).");
 *      p.solve("n(A, [cat]).");
 *
 * The 2p DCG Library is set up to handle an extra DCG operator that use regEx.
 * All acummulator lists that contain more than two elements and one Variable are assumed to be on one of three forms:
 * 1) [regExAtom, Var, regExAtom]
 * 2) [regExAtom, Var]
 * 3) [Var, regExAtom]
 * 3) [regExAtom]
 *
 * The regex placed before Var in 1) and 2) is assumed to be a match to the head of the inList.
 * The inList is matched against this regex and if it starts with the regex, the match is removed.
 * The regex behind the Var in 1) and 3) is assumed to mark the end of the search.
 * The Var is everything between the two regex (1), after the regex (2), or before the regex (3).
 * Remember in case 2) that Var consumes the rest and nothing should be left in the second DCG list.
 *
 * The following examples illustrate how this DCG Library can be used to parse /*comments.
 *
 * a. using a normal DCG logic, you can parse star-comment as this (no Variables is present in the list):
 *   star_comment(CommentText) --> ['/','*'], star_comment_body(CommentText).
 *   star_comment_body([]) -->  ['*', '/'].
 *   star_comment_body([A|Out]) -->  [A], star_comment_body(Out).
 *
 *
 * b. using the regexDCGclause, you can parse star-comment as this:
 *    star_comment(CommentText) -*-> ['/', '*'], star_comment_body(CommentText).
 *    star_comment_body(CommentText) -*-> [CommentText, '\\*[/]'].                     obs: [/] can be replaced with just /
 *
 *
 * c. using the stuck_in_the_middle structure, you can parse star-comment as this:
 *    star_comment(CommentText) -*-> ['/\\*', CommentText, '\\*[/]'].
 *
 * e. using the stuck_in_the_middle structure, you can parse star-comment as this:
 *    star_comment(CommentText) -**-> ['/\\*', CommentText, '\\*[/]'].
 *
 * CommentText = "/*the text* /"
 * morph(Sign) -**-> [Sign, '\\.\s\\*'].
 *
 * d. using the consume_the_rest structure, you can box in an incomplete star-comment as this:
 *    incomplete_star_comment(TheRestOfTheText) -*-> ['/\\*', TheRestOfTheText].
 *
 * The extended DCG logic of examples b), c) and d) are detected every time a single Variable is placed
 * prior to, between Structs or after the other nonVar elements in the DCG accumulator list.
 * See examples b),c) and d) above. Only a single variable can be placed in the DCG accumulator lists.
 *
 * @author ivar.orstavik@hist.no
 */
public class DCGLibraryIvar extends Library {

    public static boolean dcgFlag = false;

    private int dcg_counter;
    private int existingVarCount;

    public String getTheory(){
        DCGLibraryIvar.dcgFlag = true;
        return
                ":- op( 1200, xfx,  '-->'). \n" +
                ":- op( 1200, xfx,  '-*->'). \n" +
                ":- op( 1200, xfx,  '-&->'). \n" +
                "dcg_preprocess :- parse_dcg_clauses_impl, abolish('-->'/2), abolish('-*->'/2), abolish('-&->'/2). ";
    }

    /**
     * DCG builtin magic operations
     *
     * @param outVar
     * @param inList the list of characters entering to the method
     * @param OutListTwo the list of characters remaining
     * @return true if the stopMarker sequence is found and all the arguments is of the correct type
     */
    public boolean dcg_magic_method_inclusive_4(BindingsTable bt, Var outVar, Term marker, Struct inList, Term OutListTwo){
        if (!(marker instanceof StructAtom))
            throw new InvalidTermException("bug in DCG");

        String body = DCGLibraryIvar.listToString(bt, inList);

        String start = Parser.removeApices(marker.toString());
        String mightBeTail = body.replaceFirst(start, "");
        if (body.endsWith(mightBeTail)){
            String head = body.substring(0, body.length()- mightBeTail.length());
            bt.unify(outVar,BindingsTable.wrapWithID(Parser.create(head), bt.getUniqueExecutionCtxID()));
            bt.unify(OutListTwo,BindingsTable.wrapWithID(Parser.create(mightBeTail), bt.getUniqueExecutionCtxID()));
            return true;
        }
        return false;
    }

    public boolean dcg_magic_method_5(BindingsTable bt, Var outVar, Term startM, Term stopM, Struct inList, Term OutListTwo){

        String body = DCGLibraryIvar.listToString(bt, inList);

        if (!(startM == Struct.emptyList)) {
            String start = Parser.removeApices(startM.toString());
            String[] firstSplit = body.split(start, 2);
            if (firstSplit[0].equals("") && firstSplit.length ==2)
                body = firstSplit[1];
            else
                return false;
        }
        if (!(stopM == Struct.emptyList)) {
            String stop = Parser.removeApices(stopM.toString());
            String[] secondSplit = body.split(stop, 2);
            if (secondSplit.length ==2) {

                bt.unify(outVar,BindingsTable.wrapWithID(Parser.create(secondSplit[0]), bt.getUniqueExecutionCtxID()));
                bt.unify(OutListTwo,BindingsTable.wrapWithID(Parser.create(secondSplit[1]), bt.getUniqueExecutionCtxID()));
                return true;
            }
            return false;
        }
        bt.unify(outVar,BindingsTable.wrapWithID(Parser.create(body), bt.getUniqueExecutionCtxID()));
        bt.unify(OutListTwo,Struct.emptyList);
        return true;
    }

    private static String listToString(BindingsTable bt, Struct inStructList) {
        StringBuffer stopSigns = new StringBuffer();
        for (Iterator it = bt.structListIterator(inStructList, true); it.hasNext();) {
            String character = it.next().toString();
            if (character.startsWith("'") && character.endsWith("'") && character.length()!=1)
                character = character.substring(1,character.length()-1);
            stopSigns.append(character);
        }
        return stopSigns.toString();
    }

    public boolean parse_dcg_clauses_impl_0(BindingsTable bt) throws PrologException {
        try {
            for (Iterator dcgs = engine.find("'-->'/2").iterator(); dcgs.hasNext();) {
                Struct dcg = (Struct) dcgs.next();
                Struct converted = (Struct) convert(dcg, 0);
                engine.assertZ(converted);
            }
        } catch (PrologException e) {
        }
        try {
            for (Iterator dcgs = engine.find("'-*->'/2").iterator(); dcgs.hasNext();) {
                Struct dcg = (Struct) dcgs.next();
                Struct converted = (Struct) convert(dcg, 1);
                engine.assertZ(converted);
            }
        } catch (PrologException e) {
        }
        try {
            for (Iterator dcgs = engine.find("'-&->'/2").iterator(); dcgs.hasNext();) {
                Struct dcg = (Struct) dcgs.next();
                Struct converted = (Struct) convert(dcg, 2);
                engine.assertZ(converted);
            }
        } catch (PrologException e) {
        }
        return true;
    }


    private synchronized Term convert(Struct t, int regEx) throws InvalidTermException {
        //initialize new ids for accumulator
        dcg_counter = 0;
        existingVarCount = new BindingsTable().getChildrenVars(t).size() +1;
        Struct s = (Struct) t.getArg(0);


        //this is a dcg clause
        //1 body workout
        Term newB = convertBody((Struct) s.getArg(1), regEx);

        //2 head case
        Term h = s.getArg(0);
        if (!(h instanceof Struct))
            throw new InvalidTermException("DCG FUCK UP on head");

        Struct oldH = (Struct) h;

        //1b new head case argument update
        LinkedList newArgs = new LinkedList();
        for (int i = 0; i < oldH.getArity(); i++)
            newArgs.add(oldH.getArg(i));
        newArgs.add(new Var("DCG_P_0", existingVarCount + 0));
        newArgs.add(new Var("DCG_P_" + dcg_counter, existingVarCount + dcg_counter));

        Struct newH = new Struct(oldH.getName(), (Term[]) newArgs.toArray(new Term[0]));

        //3 return the finished product
        return new Struct(":-", newH, newB);
    }

    private Term convertBody(Struct t, int regEx) throws InvalidTermException {
        //if t is a {...}, then just add the body of the {} as the child
        if ("{}".equals(t.getName()))
            return t.getArg(0);

        //if t is an operator, don't change it, but go recursivly into it and dcg it's children
        if (t.opType != null || t.hasPredicateIndicator("!",0) || t.hasPredicateIndicator(",",2)){
            LinkedList newArgs = new LinkedList();
            for (int i = 0; i < t.getArity(); i++) {
                Term child = t.getArg(i);
                if (child instanceof Struct)
                    child = convertBody((Struct) child, regEx);
                newArgs.add(child);
            }
            return new Struct(t.getName(), (Term[]) newArgs.toArray(new Term[0]));
        }

        //if t is a regular Struct or StructAtom -> add DCG parameters.
        if (!(t instanceof StructList)) {
            LinkedList newArgs = new LinkedList();
            for (int i = 0; i < t.getArity(); i++)
                newArgs.add(t.getArg(i));
            newArgs.add(new Var("DCG_P_" + dcg_counter, existingVarCount + dcg_counter));
            newArgs.add(new Var("DCG_P_" + ++dcg_counter, existingVarCount + dcg_counter));
            return new Struct(t.getName(), (Term[]) newArgs.toArray(new Term[0]));
        }

        //t is a StructList
        Struct accumulator = (Struct) t;
        //if t is an empty list, return it.
        if (accumulator == Struct.emptyList)
            return accumulator;

        if (!BasicLibrary.list_1(new BindingsTable(), accumulator))
            throw new RuntimeException("What to do with [a|b] in DCG clauses?");

        //else t is a list and not empty    [...]

        Var dcg0 = new Var("DCG_P_" + dcg_counter, existingVarCount + dcg_counter);
        Var dcg1 = new Var("DCG_P_" + ++dcg_counter, existingVarCount + dcg_counter);

        LinkedList terms = new LinkedList();
        for (Iterator it = new BindingsTable().structListIterator(accumulator, true); it.hasNext();)
            terms.add(it.next());
        //ordinary DCG -->
        if (regEx == 0){
            terms.add(dcg1);    //adds ..|dcg1]
            return new Struct("=", dcg0, Struct.createStructList(terms));
        }

        //inclusive regex DCG -&->
        if (regEx == 2){
            Var var = (Var) terms.get(0);
            StructAtom marker = (StructAtom) terms.get(1);
            return new Struct("dcg_magic_method_inclusive", var, marker, dcg0, dcg1);
        }

        //exclusive regex DCG -*->
//        LinkedList terms = new LinkedList();
//        for (Iterator it = new BindingsTable().structListIterator(accumulator); it.hasNext();)
//            terms.add(it.next());
        if (terms.size() == 3)
            return new Struct("dcg_magic_method", (Var)terms.get(1), (StructAtom)terms.get(0), (StructAtom)terms.get(2), dcg0, dcg1);
        if (terms.size() == 2){
            Term arg0 = (Term) terms.get(0);
            Term arg1 = (Term) terms.get(1);
            if (arg0 instanceof Var)
                return new Struct("dcg_magic_method", arg0, Struct.emptyList, arg1, dcg0, dcg1);
            return new Struct("dcg_magic_method", arg1, arg0, Struct.emptyList, dcg0, dcg1);
        }
        if (terms.size() == 1)
            return new Struct("dcg_magic_method", new Var("_", dcg_counter + ++existingVarCount), (StructAtom)terms.get(0), Struct.emptyList, dcg0, dcg1);
        throw new InvalidTermException("invalid DCG -*-> term: " + accumulator +
                "\nLook at the javadoc for the DCGLibrary class and the DCGLibraryTestCase.");
    }
}
