package jTrolog.parser;

import jTrolog.errors.InvalidTermException;
import jTrolog.engine.*;
import jTrolog.terms.Number;
import jTrolog.terms.Double;
import jTrolog.terms.*;

import java.io.*;
import java.util.*;
import java.util.regex.Pattern;

/**
 * This class defines a parser of prolog terms and sentences.
 * <p/>
 * BNF for jTrolog part 2: Parser
 * term ::= exprA(1200)
 * exprA(n) ::= exprB(n) { op(yfx,n) exprA(n-1) |
 *                         op(yf,n) }*
 * exprB(n) ::= exprC(n-1) { op(xfx,n) exprA(n-1) |
 *                           op(xfy,n) exprA(n) |
 *                           op(xf,n) }*
 * // exprC is called parseLeftSide in the code
 * exprC(n) ::= '-' integer | '-' float |
 *              op( fx,n ) exprA(n-1) |
 *              op( fy,n ) exprA(n) |
 *              exprA(n)
 * exprA(0) ::= integer |
 *              float |
 *              atom |
 *              variable |
 *              atom'(' exprA(1200) { ',' exprA(1200) }* ')' |
 *              '[' [ exprA(1200) { ',' exprA(1200) }* [ '|' exprA(1200) ] ] ']' |
 *              '(' { exprA(1200) }* ')'
 *              '{' { exprA(1200) }* '}'
 * op(type,n) ::= atom | { symbol }+
 */
public class Parser implements Serializable {
    int dqFlag = 0;
    private static final int DQ_ATOMS = 0;
    private static final int DQ_CHARS = 1;
    private static final int DQ_CODES = 2;

    public static Struct create(String charList){
        if (charList == null || charList.length()==0)
            return Struct.emptyList;

        LinkedList elements = new LinkedList();
        for (int i = 0; i < charList.length(); i++)
            elements.add(new StructAtom(Character.toString(charList.charAt(i))));
        elements.add(Struct.emptyList);
        return Struct.createStructList(elements);
    }

    public static Struct create(int[] numbers){
        if (numbers == null || numbers.length == 0)
            return Struct.emptyList;

        LinkedList elements = new LinkedList();
        for (int i = 0; i < numbers.length; i++)
            elements.add(new Int(numbers[i]));
        elements.add(Struct.emptyList);
        return Struct.createStructList(elements);
    }

    public static WrapStructList createNewList(int lengthInt, int ctxId) {
        LinkedList vars = new LinkedList();
        for (int i = 0; i < lengthInt; i++)
            vars.add(new Var("_", i));
        vars.add(Struct.emptyList);
        return (WrapStructList) BindingsTable.wrapWithID(Struct.createStructList(vars), ctxId);
    }

    public static Term createStruct(String name, Term[] argValuesToBeLinked, BindingsTable bt) {
        Term[] variableArgs = new Term[argValuesToBeLinked.length];
        for (int i = 0; i < argValuesToBeLinked.length; i++)
            variableArgs[i] = new Var("_", i);
        WrapStruct wrapStructRes = (WrapStruct) BindingsTable.wrapWithID(new Struct(name, variableArgs), bt.getUniqueExecutionCtxID());
        for (int i = 0; i < argValuesToBeLinked.length; i++)
            bt.unify(wrapStructRes.getArg(i), argValuesToBeLinked[i]);
        return wrapStructRes;
    }

    private static class IdentifiedTerm {
        private int priority;
        private Term result;
        public IdentifiedTerm(String opType, int priority, Term result) {
            this.priority = priority;
            this.result = result;
            if (opType != null){
                ((Struct) result).opType = opType;
                ((Struct) result).opPrio = (short) priority;
            }
        }
    }

    private Tokenizer tokenizer;
    private Prolog opTable;
    private ArrayList variableMap;

    /**
     * creating a Parser specifing how to handle operators
     * and what text to parse
     */
    public Parser(InputStream theoryText, Prolog p) {
        this(p, new Tokenizer(new BufferedReader(new InputStreamReader(theoryText))));
    }

    /**
     * creating a Parser specifing how to handle operators
     * and what text to parse
     */
    public Parser(String theoryText, Prolog engine) {
        this(engine, new Tokenizer(theoryText));
    }

    /**
     * creating a parser with default operator interpretation
     */
    public Parser(String theoryText) {
        this(null, new Tokenizer(theoryText));
    }

    /**
     * creating a parser with default operator interpretation
     */
    public Parser(InputStream theoryText) {
        this(null, new Tokenizer(new BufferedReader(new InputStreamReader(theoryText))));
    }

    private Parser(Prolog p, Tokenizer lexer){
        tokenizer = lexer;
        variableMap = new ArrayList();
        if (p == null) {
            opTable = Prolog.defaultMachine;
        } else {
            opTable = p;
            Term dqFlag = p.getFlagValue("double_quotes");
            if (dqFlag != null){
                if ("chars".equals(dqFlag.toString()))
                    this.dqFlag = DQ_CHARS;
                else if ("codes".equals(dqFlag.toString()))
                    this.dqFlag = DQ_CODES;
                else
                    this.dqFlag = DQ_ATOMS;
            }
        }
    }

//  user interface

    public Iterator iterator() throws InvalidTermException {
        return new TermIterator(this);
    }

    /**
     * Parses next term from the stream built on string.
     * @param endNeeded <tt>true</tt> if it is required to parse the end token
     * (a period), <tt>false</tt> otherwise.
     * @throws InvalidTermException if a syntax error is found.
     */
    public Term nextTerm(boolean endNeeded) throws InvalidTermException {
        try {
            variableMap.clear();
            Token t = tokenizer.readToken();
            if (t.isEOF())
                return null;

            tokenizer.unreadToken(t);
            Term term = expr(false);
            if (term == null)
            	throw new InvalidTermException("The parser is unable to finish.");

            if (endNeeded && ! tokenizer.readToken().isType('.'))
                throw new InvalidTermException("The term " + term + " is not ended with a period.");
            return term;
        } catch (IOException ex) {
            throw new InvalidTermException("An I/O error occured.");
        }
    }

    // internal parsing procedures

    private Term expr(boolean commaIsEndMarker) throws InvalidTermException, IOException {
        return exprA(Prolog.OP_HIGH, commaIsEndMarker).result;
    }

    private IdentifiedTerm exprA(int maxPriority, boolean commaIsEndMarker) throws InvalidTermException, IOException {

        IdentifiedTerm leftSide = exprB(maxPriority, commaIsEndMarker);

        //{op(yfx,n) exprA(n-1) | op(yf,n)}*
        Token t = tokenizer.readToken();
        for (; t.isOperator(commaIsEndMarker); t = tokenizer.readToken()) {

            int YFX = opTable.getOperatorPriority(t.seq, "yfx");
            int YF = opTable.getOperatorPriority(t.seq, "yf");

            //YF and YFX has a higher priority than the left side expr and less then top limit
            // if (YF < leftSide.priority && YF > OperatorTable.OP_HIGH) YF = -1;
            if (YF < leftSide.priority || YF > maxPriority) YF = -1;
            // if (YFX < leftSide.priority && YFX > OperatorTable.OP_HIGH) YFX = -1;
            if (YFX < leftSide.priority || YFX > maxPriority) YFX = -1;

            //YFX has priority over YF
            if (YFX >= YF && YFX >= Prolog.OP_LOW){
                IdentifiedTerm ta = exprA(YFX-1, commaIsEndMarker);
                if (ta != null) {
                    leftSide = new IdentifiedTerm("YFX", YFX, Struct.createList(t.seq, leftSide.result, ta.result));
                    continue;
                }
            }
            //either YF has priority over YFX or YFX failed
            if (YF >= Prolog.OP_LOW) {
                leftSide = new IdentifiedTerm("YF", YF, new Struct(t.seq, leftSide.result));
                continue;
            }
            break;
        }
        tokenizer.unreadToken(t);
        return leftSide;
    }

    private IdentifiedTerm exprB(int maxPriority, boolean commaIsEndMarker) throws InvalidTermException, IOException {

        //1. op(fx,n) exprA(n-1) | op(fy,n) exprA(n) | expr0
        IdentifiedTerm left = parseLeftSide(commaIsEndMarker, maxPriority);

        //2.left is followed by either xfx, xfy or xf operators, parse these
        Token operator = tokenizer.readToken();
        for (; operator.isOperator(commaIsEndMarker); operator = tokenizer.readToken()) {
            int XFX = opTable.getOperatorPriority(operator.seq, "xfx");
            int XFY = opTable.getOperatorPriority(operator.seq, "xfy");
            int XF = opTable.getOperatorPriority(operator.seq, "xf");

            //check that no operator has a priority higher than permitted
            //or a lower priority than the left side expression
            if (XFX > maxPriority || XFX < Prolog.OP_LOW) XFX = -1;
            if (XFY > maxPriority || XFY < Prolog.OP_LOW) XFY = -1;
            if (XF > maxPriority || XF < Prolog.OP_LOW) XF = -1;

            //XFX
            boolean haveAttemptedXFX = false;
            if (XFX >= XFY && XFX >= XF && XFX >= left.priority) {     //XFX has priority
                IdentifiedTerm found = exprA(XFX - 1, commaIsEndMarker);
                if (found != null) {
                    Struct xfx = Struct.createList(operator.seq, left.result, found.result);
                    left = new IdentifiedTerm("XFX", XFX, xfx);
                    continue;
                } else
                    haveAttemptedXFX = true;
            }
            //XFY
            if (XFY >= XF && XFY >= left.priority){           //XFY has priority, or XFX has failed
                IdentifiedTerm found = exprA(XFY, commaIsEndMarker);
                if (found != null) {
                    Struct xfy = Struct.createList(operator.seq, left.result, found.result);
                    left = new IdentifiedTerm("XFY", XFY, xfy);
                    continue;
                }
            }
            //XF
            if (XF >= left.priority)                   //XF has priority, or XFX and/or XFY has failed
                return new IdentifiedTerm("XF", XF, new Struct(operator.seq, left.result));

            //XFX did not have top priority, but XFY failed
            if (!haveAttemptedXFX && XFX >= left.priority) {
                IdentifiedTerm found = exprA(XFX - 1, commaIsEndMarker);
                if (found != null) {
                    Struct xfx = Struct.createList(operator.seq, left.result, found.result);
                    left = new IdentifiedTerm("XFX", XFX, xfx);
                    continue;
                }
            }
            break;
        }
        tokenizer.unreadToken(operator);
        return left;
    }

    /**
     * Parses and returns a valid 'leftside' of an expression.
     * If the left side starts with a prefix, it consumes other expressions with a lower priority than itself.
     * If the left side does not have a prefix it must be an expr0.
     *
     * @param commaIsEndMarker used when the leftside is part of and argument list of expressions
     * @param maxPriority operators with a higher priority than this will effectivly end the expression
     * @return a wrapper of: 1. term correctly structured and 2. the priority of its root operator
     * @throws InvalidTermException
     */
    private IdentifiedTerm parseLeftSide(boolean commaIsEndMarker, int maxPriority) throws InvalidTermException, IOException {
        //1. prefix expression
        Token f = tokenizer.readToken();
        if (f.isOperator(commaIsEndMarker) && !f.isFunctor()) {
            int FX = opTable.getOperatorPriority(f.seq, "fx");
            int FY = opTable.getOperatorPriority(f.seq, "fy");

            if (f.seq.equals("-")) {
            	Token t = tokenizer.readToken();
            	if (t.isNumber())
                    return new IdentifiedTerm(null, 0, jTrolog.terms.Number.create("-" + t.seq));
            	else
            		tokenizer.unreadToken(t);
            }

            //check that no operator has a priority higher than permitted
            if (FY > maxPriority) FY = -1;
            if (FX > maxPriority) FX = -1;


            //FX has priority over FY
            boolean haveAttemptedFX = false;
            if (FX >= FY && FX >= Prolog.OP_LOW){
                IdentifiedTerm found = exprA(FX-1, commaIsEndMarker);    //op(fx, n) exprA(n - 1)
                if (found != null)
                    return new IdentifiedTerm("FX", FX, new Struct(f.seq, found.result));
                else
                    haveAttemptedFX = true;
            }
            //FY has priority over FX, or FX has failed
            if (FY >= Prolog.OP_LOW) {
                IdentifiedTerm found = exprA(FY, commaIsEndMarker); //op(fy,n) exprA(1200)  or   op(fy,n) exprA(n)
                if (found != null)
                    return new IdentifiedTerm("FY", FY, new Struct(f.seq, found.result));
            }
            //FY has priority over FX, but FY failed
            if (!haveAttemptedFX && FX >= Prolog.OP_LOW) {
                IdentifiedTerm found = exprA(FX-1, commaIsEndMarker);    //op(fx, n) exprA(n - 1)
                if (found != null)
                    return new IdentifiedTerm("FX", FX, new Struct(f.seq, found.result));
            }
        }
        tokenizer.unreadToken(f);
        //2. expr0
        return new IdentifiedTerm(null, 0, expr0());
    }

    /**
     * exprA(0) ::= integer |
     *              float |
     *              variable |
     *              atom |
     *              atom( exprA(1200) { , exprA(1200) }* ) |
     *              '[' exprA(1200) { , exprA(1200) }* [ | exprA(1200) ] ']' |
     *              '{' [ exprA(1200) ] '}' |
     *              '(' exprA(1200) ')'
     */
    private Term expr0() throws InvalidTermException, IOException {
        Token t1 = tokenizer.readToken();

        if (t1.isType(Token.INTEGER))
            return Int.create(t1.seq);

        if (t1.isType(Token.FLOAT))
            return new Double(t1.seq);

        if (t1.isType(Token.VARIABLE)){
            int pos = variableMap.indexOf(t1.seq);
            if (pos != -1 && t1.seq != Var.ANY)
                return new Var(t1.seq, pos+1);
            variableMap.add(t1.seq);
            return new Var(t1.seq, variableMap.size());
        }

        if (t1.isFunctor()) {
            String functor = t1.seq;
            Token t = tokenizer.readToken();   //reading left par
            LinkedList l = new LinkedList();
            do {
                l.add(expr(true));              //adding argument
                t = tokenizer.readToken();
                if (")".equals(t.seq))          //if right par, return
                    return new Struct(functor, (Term[]) l.toArray(new Term[0]));
            } while (",".equals(t.seq));        //if comma, read next arg
            throw new InvalidTermException("Error in argument list syntax.\n" +
                    "Token: " + t + " not expected at line " + tokenizer.lineno() + ".");
        }

        if (t1.isType(Token.DQ_SEQUENCE)) {
            if (dqFlag == Parser.DQ_ATOMS)  //DQ as atoms
                return Struct.createStruct(t1.seq);
            if (dqFlag == Parser.DQ_CHARS)  //DQ as char list
                return create(t1.seq);
            if (dqFlag == Parser.DQ_CODES) {  //DQ as int list
                char[] chars = t1.seq.toCharArray();
                int[] codes = new int[chars.length];
                System.arraycopy(chars, 0, codes, 0, chars.length);
                return create(codes);
            }
        }

        if (t1.isAtom())
            return Struct.createStruct(t1.seq);

        if (t1.isType('(')) {
            Term term = expr(false);
            if (tokenizer.readToken().isType(')'))
                return term;
            throw new InvalidTermException("Missing right parenthesis: ("+term + " -> here <-");
        }

        if (t1.isType('[')) {
            Token t2 = tokenizer.readToken();
            if (t2.isType(']'))
                return Struct.emptyList;
            tokenizer.unreadToken(t2);

            LinkedList elems = new LinkedList();
            do {
                elems.add(expr(true));
                t2 = tokenizer.readToken();
                if ("|".equals(t2.seq)){
                    elems.add(expr(true));
                    t2 = tokenizer.readToken();
                    if ("]".equals(t2.seq))
                        return Struct.createStructList(elems);
                    throw new InvalidTermException("Missing ']' after: "+ elems.getLast());
                }
                if ("]".equals(t2.seq)) {
                    elems.add(Struct.emptyList);
                    return Struct.createStructList(elems);
                }
            } while(",".equals(t2.seq));
            throw new InvalidTermException("Error in list syntax after: "+ elems.getLast());
        }

        if (t1.isType('{')) {
            Token t2 = tokenizer.readToken();
            if (t2.isType('}'))
                return new StructAtom("{}");

            tokenizer.unreadToken(t2);
            Term arg = expr(false);
            t2 = tokenizer.readToken();
            if (t2.isType('}'))
                return new Struct("{}", arg);
            throw new InvalidTermException("Missing right braces: {"+arg + " -> here <-");
        }

        throw new InvalidTermException("The following token could not be identified: "+t1.seq);
    }

    public int getCurrentLine() {
        return tokenizer.lineno();
    }

    /**
     * @return true if the String could be a prolog atom
     */
    public static boolean isAtom(String s) {
        return atom.matcher(s).matches();
    }

    static private Pattern atom = Pattern.compile("(!|[a-z][a-zA-Z_0-9]*)");

    public static Number parseNumber(String s) throws InvalidTermException {
        Term t = new Parser(s).nextTerm(false);
        if (t instanceof Number)
            return (Number) t;
        throw new InvalidTermException("Term " + t + " is not a number.");
    }

    public static String removeApices(String st){
        if (st.startsWith("'")&&st.endsWith("'") && st.length() > 2)
            return st.substring(1,st.length()-1);
       return st;
    }
}