package jTrolog.terms;

import jTrolog.parser.Parser;
import jTrolog.terms.EvaluableTerm;
import jTrolog.engine.BindingsTable;

import java.util.*;

/**
 * Struct class represents both compound prolog term
 * and atom term (considered as 0-arity compound).
 */
public class Struct implements EvaluableTerm {

    private String name;
    private Term[] args;
    private String predicateIndicator;  /** speedup hash map operation */

    public String opType;
    public short opPrio = -1;
    private Var[] varList;

    public static Struct emptyList = new StructEmptyList();

    public Var[] getVarList() {
        return varList;
    }

    private static class StructEmptyList extends Struct {
        StructEmptyList() {
            super("[]", new Term[0]);
        }
        public String toString() {
            return "[]";
        }
    }


    public final static String listSignature = "'.'/2".intern();
    int size = 0;
    Term last;

    Struct() {
    }

    /**
     * Builds a compound, with one argument
     */
    public Struct(String f,Term at0) {
        this(f,new Term[]{at0});
    }

    /**
     * Builds a compound, with two arguments
     */
    public Struct(String f, Term at0, Term at1) {
        this(f,new Term[]{at0, at1});
    }

    /**
     * Builds a compound, with three arguments
     */
    public Struct(String f,Term at0,Term at1,Term at2) {
        this(f,new Term[]{at0, at1, at2});
    }

    /**
     * Builds a compound, with four arguments
     */
    public Struct(String f,Term at0,Term at1,Term at2,Term at3) {
        this(f,new Term[]{at0, at1, at2, at3});
    }

    /**
     * Builds a compound, with five arguments
     */
    public Struct(String f,Term at0,Term at1,Term at2,Term at3,Term at4) {
        this(f,new Term[]{at0, at1, at2, at3, at4});
    }

    /**
     * Builds a compound, with six arguments
     */
    public Struct(String f,Term at0,Term at1,Term at2,Term at3,Term at4,Term at5) {
        this(f,new Term[]{at0, at1, at2, at3, at4, at5});
    }

    /**
     * Builds a compound, with seven arguments
     */
    public Struct(String f,Term at0,Term at1,Term at2,Term at3,Term at4,Term at5,Term at6) {
        this(f,new Term[]{at0, at1, at2, at3, at4, at5, at6});
    }

    /**
     * Builds a compound, with an array of arguments
     */
    public Struct(String f,Term[] argList) {
        this(f,argList.length);
        args = argList;
        List childrenVarList = getChildrenVarList(args);
        if (childrenVarList != null)
            varList = (Var[]) childrenVarList.toArray(new Var[0]);
    }

    private Struct(String n, int arity) {
        name = (Parser.removeApices(n)).intern();
        predicateIndicator = ((Parser.isAtom(name) ? name : "'"+ name + "'") + "/" + arity).intern();
    }

    /**
     * only used by BindingsTable flatclone
     * @param toBeCopied
     * @param elems
     */
    public Struct(Struct toBeCopied, Term[] elems) {
        opType = toBeCopied.getOpType();
        opPrio = (short) toBeCopied.getOpPrio();
        name = toBeCopied.getName();
        predicateIndicator = toBeCopied.getPredicateIndicator();
        args = elems;
    }

    public String getPredicateIndicator() {
        return predicateIndicator;
    }

    public boolean hasPredicateIndicator(String name, int arity) {
        if (getArity() != arity)
            return false;
        name = Parser.removeApices(name);
        return name.equals(getName());
    }

    /**
     * Gets the number of elements of
     * this structure
     */
    public int getArity() {
        return args.length;
    }

    /**
     * Gets the functor name  of this structure
     */
    public String getName() {
        return name;
    }

    /**
     * Gets the i-th element of this structure
     *
     * No bound check is done
     */
    public Term getArg(int index) {
        return args[index];
    }

    /**
     * Sets the i-th element of this structure
     *
     * (Only for internal service)
     */
    public void setArg(int index, Term argument) {
        args[index] = argument;
    }

    /**
     * Gets the string representation of this structure
     *
     * Specific representations are provided for lists and atoms.
     * Names starting with upper case letter are enclosed in apices.
     */
    public String toString() {
        if (this instanceof StructList)
            return toStringList();
        return toStringImpl(false);
    }

    public String toStringList() {
        StringBuffer res = new StringBuffer("[");
        boolean firstPass = true;
        for (Iterator it = iterator((StructList) this); it.hasNext();) {
            Term child = (Term) it.next();
            if (!it.hasNext()) { //last child
                if (child == emptyList)
                    return res.append(']').toString();
                return res.append('|').append(child).append(']').toString();
            }
            if (firstPass){
                res.append(child);
                firstPass = false;
            } else
                res.append(", ").append(child);
        }
        throw new RuntimeException("bug");
    }

    public String toStringListSmall() {
        StringBuffer res = new StringBuffer("[");
        boolean firstPass = true;
        int stopPrinting = 10;
        for (Iterator it = iterator((StructList) this); it.hasNext();) {
            if (stopPrinting-- == 0){
                Term last = getLastChild(((StructList) this));
                if (last == emptyList)
                    return res.append(" ... ").append("]").toString();
                return res.append(" ... ").append("|").append(last).append("]").toString();
            }
            Term child = (Term) it.next();
            if (!it.hasNext()) { //last child
                if (child == emptyList)
                    return res.append(']').toString();
                return res.append('|').append(child).append(']').toString();
            }
            if (firstPass){
                res.append(child);
                firstPass = false;
            } else
                res.append(", ").append(child);
        }
        throw new RuntimeException("bug");
    }

    public String toStringSmall(){
        if (this instanceof StructList)
            return toStringListSmall();
        return toStringImpl(true);
    }

    private String toStringImpl(boolean small) {
        String opType = getOpType();
        String name = getName();
        int opPrio = getOpPrio();

        if(name == null)
            System.out.println("wtf");
        if (name.equals("{}")) {
            if (getArity() == 0)
                return name;
            if (getArity() == 1) {
                Term arg = getArg(0);
                return "{" + (small ? arg.toStringSmall() : arg.toString()) + "}";
            }
        }
        //todo a bit hackish
        if (opType == null && (hasPredicateIndicator(":-", 2) || hasPredicateIndicator(",", 2))){
            opType ="XFX";
            opPrio = 1200;
        }
        if (opType != null) {
            int onePrioMax = opType.equals("YF") || opType.equals("FY") || opType.equals("YFX") ? opPrio : opPrio -1;
            int twoPrioMax = opType.equals("XFY") ? opPrio : opPrio -1;

            Term childOne = getArg(0);
            String a = small ? childOne.toStringSmall() : childOne.toString();
            if (childOne instanceof Struct && ((Struct) childOne).opPrio > onePrioMax)
                a = "(" + a + ")";
            if (opType.equals("XF") || opType.equals("YF"))
                return a + name;
            else if (opType.equals("FX") || opType.equals("FY"))
                return name + a;

            Term childTwo = getArg(1);
            String b = small ? childTwo.toStringSmall() : childTwo.toString();
            if (childTwo instanceof Struct && ((Struct) childTwo).opPrio > twoPrioMax)
                b = "(" + b + ")";
            if (hasPredicateIndicator(",",2))
                return a + name + " " + b;
            return a + " " + name + " " + b;
        }
        StringBuffer res = new StringBuffer(Parser.isAtom(name) ? name : "'" + name + "'");
        for (int i = 0;i < getArity();i++) {
            if (i == 0)
                res.append("(");
            Term arg = getArg(i);
            if (arg instanceof Struct && ((Struct) arg).hasPredicateIndicator(",", 2))
                res.append("(").append(small ? arg.toStringSmall() : arg.toString()).append(")");
            else
                res.append(small ? arg.toStringSmall() : arg.toString());
            res.append(i < getArity()-1 ? ", ": ")");
        }
        return res.toString();
    }

    int getOpPrio() {
        return opPrio;
    }

    String getOpType() {
        return opType;
    }

    /**
     * Check is this struct is clause or directive
     * @param struct
     */
    public static boolean isClause(Struct struct) {
        return ":-".equals(struct.getName()) && struct.getArity() > 1 && struct.getArg(0) instanceof Struct;
    }

    //todo might be problematic use of
    public static Struct createList(String f, Term at0, Term at1) {
        return ".".equals(f) ? createStructList(new LinkedList(Arrays.asList(new Term[]{at0, at1}))) :  new Struct(f, at0, at1);
    }

    public static Struct createStruct(String f) {
    	if ("[]".equals(f))
    		return emptyList;
    	return new StructAtom(f);
    }

    public Iterator children(){
        return new Iterator() {
            int pos = 0;

            public boolean hasNext() {
                return pos < args.length;
            }

            public Object next() {
                return getArg(pos++);
            }

            public void remove() {
                throw new IllegalAccessError("remove is an illegal operation on the children iterator of Structs");
            }
        };
    }

    private static List getChildrenVarList(Term[] args) {
        ArrayList findings = new ArrayList();
        for (int i = 0; i < args.length; i++) {
            Term arg = args[i];
            if (arg instanceof Var)
                findings.add(arg);
            else if (arg instanceof Struct && ((Struct) arg).getVarList() != null){
                for (int j = 0; j < ((Struct) arg).getVarList().length; j++)
                    findings.add(((Struct) arg).getVarList()[j]);
            }
        }
        return findings.isEmpty() ? null : findings;
    }

    public boolean equals(Object t) {
        if (!(t instanceof Struct))
            return false;
        Struct other = (Struct) t;
        if (!getPredicateIndicator().equals(other.getPredicateIndicator()))
            return false;
        Iterator it = children();
        Iterator it2 = other.children();
        while(it.hasNext()){
            if (!it2.hasNext())
                return false;
            Term myChild = (Term) it.next();
            Term otherChild = (Term) it2.next();
            if (!myChild.equals(otherChild))
                return false;
        }
        return true;
    }

    public static Struct createStructList(LinkedList complete) {
        if (complete.isEmpty())
            return emptyList;
        if (complete.size() == 2)
            return new StructList((Term)complete.getFirst(),(Term)complete.getLast());
        if (complete.size() > 2){
            Term head = (Term) complete.removeFirst();
            return new StructList(head, createStructList(complete));
        }
        throw new RuntimeException("omg-..");
    }

    public static Iterator iterator(StructList structList){
        final Struct origin = (Struct) (structList instanceof Wrapper ? ((Wrapper) structList).getBasis() : structList);
        Iterator it = new Iterator() {
            Term hereIAmBaby = origin;

            public boolean hasNext() {
                return hereIAmBaby != null;
            }

            public Object next() {
                if (hereIAmBaby == null)
                    throw new IndexOutOfBoundsException("iterating out of list");
                if (hereIAmBaby == emptyList) {
                    hereIAmBaby = null;
                    return emptyList;
                }
                if (hereIAmBaby instanceof Struct && ((Struct) hereIAmBaby).hasPredicateIndicator(".", 2)) {
                    Term res = ((Struct) hereIAmBaby).getArg(0);
                    hereIAmBaby = ((Struct) hereIAmBaby).getArg(1);
                    return res;
                }
                Term res = hereIAmBaby;
                hereIAmBaby = null;
                return res;
            }

            public void remove() {
                throw new UnsupportedOperationException("don't delete on List iteration");
            }
        };
        if (structList instanceof Wrapper)
            return new Wrapper.WrappedIterator(it, ((Wrapper) structList).getContext());
        return it;
    }

    public static int size(StructList structList) {
        if (structList instanceof Wrapper)
            return ((StructList) ((Wrapper) structList).getBasis()).size;
        return structList.size;

    }

    public static Term getLastChild(StructList structList) {
        if (structList instanceof Wrapper)
            return BindingsTable.wrapWithID(((StructList) ((Wrapper) structList).getBasis()).last, ((Wrapper) structList).getContext());
        return structList.last;
    }

    public boolean isPossiblyList() {
        return predicateIndicator == listSignature;
    }
}