package jTrolog.engine;

import junit.framework.TestCase;
import jTrolog.errors.InvalidTermException;
import jTrolog.parser.Parser;
import jTrolog.terms.*;

public class ParserTestCase extends TestCase {

    public void testReadingTerms() throws InvalidTermException {
        Parser p = new Parser("hello.");
        Struct result = new StructAtom("hello");
        assertEquals(result, p.nextTerm(true));
    }

    public void testReadingEOF() throws InvalidTermException {
        Parser p = new Parser("");
        assertNull(p.nextTerm(false));
    }

    public void testUnaryPlusOperator() {
        Parser p = new Parser("n(+100).\n");
        // SICStus Prolog interprets "n(+100)" as "n(100)"
        // GNU Prolog interprets "n(+100)" as "n(+(100))"
        // The ISO Standard says + is not a unary operator
        try {
            assertNotNull(p.nextTerm(true));
            fail();
        } catch (InvalidTermException e) {}
    }

    public void testUnaryMinusOperator() throws InvalidTermException {
        Parser p = new Parser("n(-100).\n");
        // TODO Check the interpretation by other engines
        // SICStus Prolog interprets "n(+100)" as "n(100)"
        // GNU Prolog interprets "n(+100)" as "n(+(100))"
        // What does the ISO Standard say about that?
        Struct result = new Struct("n", new Int(-100));
        assertEquals(result, p.nextTerm(true));
    }

    public void testBinaryMinusOperator() throws InvalidTermException {
        String s = "abs(3-11)";
        Parser p = new Parser(s);
        Struct result = new Struct("abs", new Struct("-", new Int(3), new Int(11)));
        assertEquals(result, p.nextTerm(false));
    }

    /**
     * see 6.3.5 in specification
     * @throws InvalidTermException
     */
    public void testListWithTail() throws InvalidTermException {
        String one = "[a]";
        String two = "[a,b]";
        String three = "[a|b]";
        String four = "[a,b|c]";
        String five= "[a|b-c]";
        String six= "[a|b,c]";

        Term t1a = new Parser(one).nextTerm(false);
        Term t1b = Struct.createList(".", new StructAtom("a"), Struct.emptyList);
        assertEquals(t1a,t1b);

        Term t2a = new Parser(two).nextTerm(false);
        Term t2b = Struct.createList(".", new StructAtom("a"), Struct.createList(".", new StructAtom("b"), Struct.emptyList));
        assertEquals(t2a,t2b);

        Term t3a = new Parser(three).nextTerm(false);
        Term t3b = Struct.createList(".", new StructAtom("a"), new StructAtom("b"));
        assertEquals(t3a,t3b);

        Term t4a = new Parser(four).nextTerm(false);
        Term t4b = Struct.createList(".", new StructAtom("a"), Struct.createList(".", new StructAtom("b"), new StructAtom("c")));
        assertEquals(t4a,t4b);

        try {
            Term t5a = new Parser(five).nextTerm(false);
            Term t5b = Struct.createList(".", new StructAtom("a"), new Struct("-", new StructAtom("b"), new StructAtom("c")));
            assertEquals(t5a,t5b);
        } catch (InvalidTermException e) {
        }

        try {
            Term t6a = new Parser(six).nextTerm(false);
            Term t6b = Struct.createList(".", new StructAtom("a"), new Struct(",", new StructAtom("b"), new StructAtom("c")));
            assertEquals(t6a,t6b);
        } catch (InvalidTermException e) {
        }

        Parser p = new Parser("[p|Y]");
        Struct result = Struct.createList(".", new StructAtom("p"), new Var("Y",1));
        assertEquals(result, p.nextTerm(false));
    }

    public void testBraces() throws InvalidTermException {
        String s = "{a, b, [3, {4, c}, 5], {a, b}}";
        Parser parser = new Parser(s);
        assertEquals(s, parser.nextTerm(false).toString());
    }

    public void testDotOperator() throws InvalidTermException {
        String s = "class('java.lang.Integer').'MAX_VALUE'";
        Prolog eng = new Prolog();
        eng.opNew(".", "xfx", 600);
        Parser p = new Parser(s, eng);
        Struct result = Struct.createList(".", new Struct("class", new StructAtom("java.lang.Integer")),
                                        new StructAtom("MAX_VALUE"));
        assertEquals(result, p.nextTerm(false));
    }

    public void testBracketedOperatorAsTerm() throws InvalidTermException {
        String s = "u (b1) b2 (b3)";
        Prolog eng = new Prolog();
        eng.opNew("u", "fx", 200);
        eng.opNew("b1", "yfx", 400);
        eng.opNew("b2", "yfx", 500);
        eng.opNew("b3", "yfx", 300);
        Parser p = new Parser(s, eng);
        Struct result = Struct.createList("b2", new Struct("u", new StructAtom("b1")), new StructAtom("b3"));
        assertEquals(result, p.nextTerm(false));
    }

    public void testBracketedOperatorAsTerm2() throws InvalidTermException {
        String s = "(u) b1 (b2) b3 a";
        Prolog eng = new Prolog();
        eng.opNew("u", "fx", 200);
        eng.opNew("b1", "yfx", 400);
        eng.opNew("b2", "yfx", 500);
        eng.opNew("b3", "yfx", 300);
        Parser p = new Parser(s, eng);
        Struct result = Struct.createList("b1", new StructAtom("u"), new Struct("b3", new StructAtom("b2"), new StructAtom("a")));
        Term actual = p.nextTerm(false);
        assertEquals(result, actual);
        assertEquals(s, actual.toString());
    }

    public void testIntegerBinaryRepresentation() throws InvalidTermException {
        String n = "0b101101";
        Parser p = new Parser(n);
        jTrolog.terms.Number result = new Int(45);
        assertEquals(result, p.nextTerm(false));
        String invalid = "0b101201";
        try {
            new Parser(invalid).nextTerm(false);
            fail();
        } catch (InvalidTermException expected) {}
    }

    public void testIntegerOctalRepresentation() throws InvalidTermException {
        String n = "0o77351";
        Parser p = new Parser(n);
        jTrolog.terms.Number result = new Int(32489);
        assertEquals(result, p.nextTerm(false));
        String invalid = "0o78351";
        try {
            new Parser(invalid).nextTerm(false);
            fail();
        } catch (InvalidTermException expected) {}
    }

    public void testIntegerHexadecimalRepresentation() throws InvalidTermException {
        String n = "0xDECAF";
        Parser p = new Parser(n);
        jTrolog.terms.Number result = new Int(912559);
        assertEquals(result, p.nextTerm(false));
        String invalid = "0xG";
        try {
            new Parser(invalid).nextTerm(false);
            fail();
        } catch (InvalidTermException expected) {}
    }

    public void testMultilineComments() throws InvalidTermException {
        String theory = "t1." + "\n" +
                "/*" + "\n" +
                "t2" + "\n" +
                "*/" + "\n" +
                "t3." + "\n";
        Parser p = new Parser(theory);
        assertEquals(new StructAtom("t1"), p.nextTerm(true));
        assertEquals(new StructAtom("t3"), p.nextTerm(true));
    }

    public void testDoubleQuotes() throws InvalidTermException {
        String theory = "t1." + "\n" +
                "a(\"go go gozilla\")." + "\n" +
                "t3." + "\n";
        Parser p = new Parser(theory);
        assertEquals(new StructAtom("t1"), p.nextTerm(true));
        assertEquals(new Struct("a", new StructAtom("go go gozilla")), p.nextTerm(true));
        assertEquals(new StructAtom("t3"), p.nextTerm(true));
    }

    public void testOperatorAsFunctor() throws InvalidTermException {
        String theory = "\\+(!).";
        Parser p = new Parser(theory);
        Term t = p.nextTerm(true);
        assertEquals(new Struct("\\+", new StructAtom("!")), t);
    }

    public void testDoubleBackSlashInQuote() throws InvalidTermException {
        String theory = "'\\\\='(a).";
        Parser p = new Parser(theory);
        Term t = p.nextTerm(true);
        assertEquals(new Struct("\\=", new StructAtom("a")), t);
    }

    // :-op(500, yfx, v). 3v2 NOT CORRECT, 3 v 2 CORRECT
    // 3+2 CORRECT, 3 + 2 CORRECT
    public void testOperatorWithNoSurroundingBlankSpace() {
        try {
            new Parser("bagof(X, Y^((X=1, Y=1) ; (X=2, Y=2)), S).").nextTerm(true);
        } catch (InvalidTermException e) {
            fail();
        }

    }

    // TODO More tests on Parser

    // Character code for Integer representation


    // +(2, 3) is now acceptable
    // what about f(+)

    // Test for {} and {...} terms

}
