package jTrolog.engine;

import junit.framework.TestCase;

import java.util.LinkedList;
import java.util.Arrays;

import jTrolog.errors.InvalidTermException;
import jTrolog.errors.PrologException;
import jTrolog.parser.Parser;
import jTrolog.terms.Struct;
import jTrolog.terms.StructAtom;
import jTrolog.terms.Term;
import jTrolog.terms.Var;

public class StructTestCase extends TestCase {

    public void testEmptyList() throws PrologException {
        Struct list = Struct.emptyList;
        BindingsTable bt = new BindingsTable();
        assertTrue(bt.isList(list));
        assertTrue(list == Struct.emptyList);
        assertEquals(0, bt.listSize(list));
        assertEquals("[]", list.getName());
        assertEquals(0, list.getArity());
    }

    /** Another correct method of building an empty list */
    public void testEmptyListAsSquaredStruct() throws PrologException {
        Struct emptyList = Struct.emptyList;
        BindingsTable bt = new BindingsTable();
        assertTrue(bt.isList(emptyList));
        assertTrue(emptyList == Struct.emptyList);
        assertEquals("[]", emptyList.getName());
        assertEquals(0, emptyList.getArity());
        assertEquals(0, bt.listSize(emptyList));
    }

    /** A wrong method of building an empty list */
    public void testEmptyListAsDottedStruct() throws PrologException {
        Struct notAnEmptyList = Struct.createStruct(".");
        assertFalse(new BindingsTable().isList(notAnEmptyList));
        assertFalse(notAnEmptyList == Struct.emptyList);
        assertEquals(".", notAnEmptyList.getName());
        assertEquals(0, notAnEmptyList.getArity());
    }

    /** Use dotted structs to build lists with content */
    public void testListAsDottedStruct() throws PrologException {
        Struct notAnEmptyList = Struct.createList(".", new StructAtom("a"), Struct.createList(".", new StructAtom("b"), Struct.emptyList));
        assertTrue(new BindingsTable().isList(notAnEmptyList));
        assertFalse(notAnEmptyList == Struct.emptyList);
        assertEquals(".", notAnEmptyList.getName());
        assertEquals(2, notAnEmptyList.getArity());
    }

    public void testListFromArgumentArray() {
        assertEquals(Struct.emptyList, Struct.emptyList);

        LinkedList l = new LinkedList();
        l.add(new StructAtom("a"));
        l.add(new StructAtom("b"));
        l.add(Struct.emptyList);
        Struct list = Struct.createStructList(l);
        assertEquals(Struct.emptyList, ((Struct) list.getArg(1)).getArg(1));
    }

    public void testListFromString() {
        Struct a = Parser.create("he");
        Term[] argList = new Term[]{new StructAtom("h"), new StructAtom("e"), Struct.emptyList};
        Struct b = Struct.createStructList(new LinkedList(Arrays.asList(argList)));
        assertEquals(a, b);
        a = Parser.create("hello");
        argList = new Term[]{new StructAtom("h"), new StructAtom("e"), new StructAtom("l"), new StructAtom("l"), new StructAtom("o"), Struct.emptyList};
        b = Struct.createStructList(new LinkedList(Arrays.asList(argList)));
        assertEquals(a, b);
        a.toString();
    }

    public void testListSize() throws PrologException {
        Struct list = Struct.createList(".", new StructAtom("a"),
                       Struct.createList(".", new StructAtom("b"),
                           Struct.createList(".", new StructAtom("c"), Struct.emptyList)));
        BindingsTable bt = new BindingsTable();
        assertTrue(bt.isList(list));
        assertFalse(list == Struct.emptyList);
        assertEquals(4, bt.listSize(list));
    }

    public void testToString() throws InvalidTermException {
        Struct emptyList = Struct.emptyList;
        assertEquals("[]", emptyList.toString());
        Struct s = new Struct("f", new Var("X", 0));
        assertEquals("f(X)", s.toString());
        Struct list = Struct.createList(".", new StructAtom("a"),
                  Struct.createList(".", new StructAtom("b"),
                      Struct.createList(".", new StructAtom("c"), Struct.emptyList)));
        assertEquals("[a, b, c]", list.toString());
    }

    public void testStructToString() throws InvalidTermException {
        String s = "C -> T ; B :- !, or((C, !, T), B)";
        Prolog eng = new Prolog();
        Parser p = new Parser(s, eng);
        Term actual = p.nextTerm(false);
        assertEquals(s, actual.toString());
    }

    public void testIsList() throws PrologException {
        Struct notList = Struct.createList(".", Struct.createStruct("a"), Struct.createStruct("b"));
        assertFalse(new BindingsTable().isList(notList));
    }

    public void testUnifyVariables() throws Throwable{
        Prolog eng = new Prolog();
        Solution res = eng.solve("'='(X, Y).");
        Term t = res.getBinding("X");
        String s = t.toString();
        if (!s.equals("Y"))
            fail();
    }
}
